#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import List

import torch
from torch import Tensor

from ._environment import _IS_TORCH_COMPILE_SUPPORTED


def maybe_torch_compile(*args, **kwargs):
    def decorator(f):
        if _IS_TORCH_COMPILE_SUPPORTED:
            return torch.compile(f, *args, **kwargs)
        return f

    return decorator


# TODO: if use cases for this grow, we might want to do a custom kernel
def merge_attentions_fn(outputs: List[Tensor], lse_tensors: List[Tensor]) -> Tensor:

    assert len(outputs) >= 2, "Expected at least two tensors."
    assert len(outputs) == len(
        lse_tensors
    ), "Expected number of outputs and LSE tensors to match."

    assert all(
        output.dim() == 4 and output.is_contiguous() for output in outputs
    ), "Output tensors must be rank-4 tensors with (batch, seq, heads, dim) contiguous layout."

    batch, seqlen, heads, dim = outputs[0].shape

    assert all(
        [x for x in output.shape] == [batch, seqlen, heads, dim] for output in outputs
    ), "Output tensors must match in shape."

    assert all(
        lse.dim() == 3
        and lse.is_contiguous()
        and [x for x in lse.shape] == [batch, seqlen, heads]
        for lse in lse_tensors
    ), "LSE tensors must be rank-3 tensors with (batch, seq, heads) contiguous layout, and match in shape."

    accum_type = torch.float32
    output_type = outputs[0].dtype

    lse_tensors = [lse.to(accum_type) for lse in lse_tensors]

    outputs = [output.to(accum_type) for output in outputs]

    lse_max = torch.maximum(lse_tensors[0], lse_tensors[1])
    for i in range(2, len(lse_tensors)):
        lse_max = torch.maximum(lse_max, lse_tensors[i])

    exp_diffs = [torch.exp(lse - lse_max).unsqueeze(-1) for lse in lse_tensors]

    outputs_rescaled = [
        output * exp_diff for output, exp_diff in zip(outputs, exp_diffs)
    ]

    assert all(
        [
            output_rescaled.shape == output.shape
            for output, output_rescaled in zip(outputs, outputs_rescaled)
        ]
    )

    sum_of_exps = exp_diffs[0] + exp_diffs[1]
    sum_of_outs = outputs_rescaled[0] + outputs_rescaled[1]
    for i in range(2, len(exp_diffs)):
        sum_of_exps += exp_diffs[i]
        sum_of_outs += outputs_rescaled[i]

    output = sum_of_outs / sum_of_exps

    output = output.to(output_type)

    return output


# TODO: if use cases for this grow, we might want to do a custom kernel
@maybe_torch_compile(fullgraph=True)
def merge_attentions_compile(
    outputs: List[Tensor], lse_tensors: List[Tensor]
) -> Tensor:
    return merge_attentions_fn(outputs, lse_tensors)



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from ..types import CausalArgType, DimensionType


def ceil_div_int(x: int, y: int) -> int:
    return (x + y - 1) // y


def ceil_div_tuple(X: tuple, Y: tuple) -> tuple:
    assert len(X) == len(Y)
    return tuple(ceil_div_int(x, y) for x, y in zip(X, Y))


def create_dim_from_int(na_dim: int, value: int) -> DimensionType:
    return tuple(value for _ in range(na_dim))  # type: ignore


def create_causal_arg_from_bool(na_dim: int, value: bool) -> CausalArgType:
    return tuple(value for _ in range(na_dim))  # type: ignore



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from ._environment import HAS_LIBNATTEN

from .backends import (
    get_bwd_configs_for_cutlass_blackwell_fmha,
    get_bwd_configs_for_cutlass_blackwell_fna,
    get_bwd_configs_for_cutlass_fmha,
    get_bwd_configs_for_cutlass_fna,
    get_bwd_configs_for_cutlass_hopper_fmha,
    get_bwd_configs_for_cutlass_hopper_fna,
    get_configs_for_cutlass_blackwell_fmha,
    get_configs_for_cutlass_blackwell_fna,
    get_configs_for_cutlass_fmha,
    get_configs_for_cutlass_fna,
    get_configs_for_cutlass_hopper_fmha,
    get_configs_for_cutlass_hopper_fna,
    get_configs_for_flex_fmha,
    get_configs_for_flex_fna,
)

from .context import (
    allow_flex_compile,
    allow_flex_compile_backprop,
    are_deterministic_algorithms_enabled,
    disable_flex_compile,
    disable_flex_compile_backprop,
    get_memory_usage_preference,
    is_flex_compile_allowed,
    is_flex_compile_backprop_allowed,
    is_kv_parallelism_in_fused_na_enabled,
    is_memory_usage_default,
    is_memory_usage_strict,
    is_memory_usage_unrestricted,
    set_memory_usage_preference,
    use_deterministic_algorithms,
    use_kv_parallelism_in_fused_na,
)

from .functional import attention, merge_attentions, na1d, na2d, na3d

from .modules import (
    NeighborhoodAttention1D,
    NeighborhoodAttention2D,
    NeighborhoodAttention3D,
)

from .version import __version__

__all__ = [
    "__version__",
    "NeighborhoodAttention1D",
    "NeighborhoodAttention2D",
    "NeighborhoodAttention3D",
    "are_deterministic_algorithms_enabled",
    "use_deterministic_algorithms",
    "use_kv_parallelism_in_fused_na",
    "is_kv_parallelism_in_fused_na_enabled",
    "set_memory_usage_preference",
    "get_memory_usage_preference",
    "is_memory_usage_default",
    "is_memory_usage_strict",
    "is_memory_usage_unrestricted",
    "is_flex_compile_allowed",
    "is_flex_compile_backprop_allowed",
    "allow_flex_compile",
    "allow_flex_compile_backprop",
    "disable_flex_compile",
    "disable_flex_compile_backprop",
    "get_bwd_configs_for_cutlass_fmha",
    "get_bwd_configs_for_cutlass_fna",
    "get_configs_for_cutlass_fmha",
    "get_configs_for_cutlass_fna",
    "get_configs_for_cutlass_hopper_fmha",
    "get_bwd_configs_for_cutlass_hopper_fmha",
    "get_configs_for_cutlass_hopper_fna",
    "get_bwd_configs_for_cutlass_hopper_fna",
    "get_bwd_configs_for_cutlass_blackwell_fmha",
    "get_bwd_configs_for_cutlass_blackwell_fna",
    "get_configs_for_cutlass_blackwell_fmha",
    "get_configs_for_cutlass_blackwell_fna",
    "get_configs_for_flex_fmha",
    "get_configs_for_flex_fna",
    "HAS_LIBNATTEN",
    "na1d",
    "na2d",
    "na3d",
    "attention",
    "merge_attentions",
]



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import functools
from typing import Optional, Tuple, Union

import torch
from torch import Tensor

from torch.amp import custom_bwd, custom_fwd
from torch.autograd import Function

amp_fwd = functools.partial(custom_fwd, device_type="cuda")
amp_bwd = functools.partial(custom_bwd, device_type="cuda")

from .._libnatten import blackwell_fmha_backward, blackwell_fmha_forward
from ..types import (
    CutlassBlackwellFmhaBackwardConfigType,
    CutlassBlackwellFmhaForwardConfigType,
    NoneType,
)
from ..utils import log
from ..utils.checks import fmha_tensor_checks

from .configs.checks import can_run_cutlass_blackwell_fmha
from .configs.cutlass_blackwell import (
    check_cutlass_blackwell_fmha_backward_config,
    check_cutlass_blackwell_fmha_forward_config,
)

logger = log.get_logger(__name__)


class CutlassBlackwellFmhaAutogradFn(Function):
    @staticmethod
    @amp_fwd
    def forward(
        ctx,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        scale: float,
        forward_config: CutlassBlackwellFmhaForwardConfigType,
        backward_config: CutlassBlackwellFmhaBackwardConfigType,
        run_persistent_kernel: bool,
    ) -> Tuple[Tensor, Tensor]:
        query = query.contiguous()
        key = key.contiguous()
        value = value.contiguous()
        output = torch.empty_like(query)

        logsumexp = torch.empty(
            query.shape[:-1], dtype=torch.float32, device=query.device
        )

        q_tile_size, kv_tile_size = forward_config
        blackwell_fmha_forward(
            output,
            query,
            key,
            value,
            logsumexp,
            scale,
            q_tile_size,
            kv_tile_size,
            run_persistent_kernel,
        )

        ctx.save_for_backward(query, key, value, logsumexp, output)
        ctx.scale = scale
        ctx.backward_config = backward_config

        return output, logsumexp

    @staticmethod
    @amp_bwd
    def backward(ctx, grad_out: Tensor, grad_lse: Tensor) -> Tuple[
        Tensor,
        Tensor,
        Tensor,
        NoneType,
        NoneType,
        NoneType,
        NoneType,
    ]:
        query, key, value, logsumexp, output = ctx.saved_tensors
        d_output = grad_out.contiguous()  # noqa: F841

        q_tile_size, k_tile_size = ctx.backward_config

        seqlen_Q = query.shape[1]

        if seqlen_Q % q_tile_size != 0:
            padding = q_tile_size - (seqlen_Q % q_tile_size)
            old_shape = query.shape
            query = torch.nn.functional.pad(
                query, (0, 0, 0, 0, 0, padding), "constant", 0
            )
            output = torch.nn.functional.pad(
                output, (0, 0, 0, 0, 0, padding), "constant", 0
            )
            d_output = torch.nn.functional.pad(
                d_output, (0, 0, 0, 0, 0, padding), "constant", 0
            )
            logsumexp = torch.nn.functional.pad(
                logsumexp, (0, 0, 0, padding), "constant", 0
            )
            seqlen_Q_aligned = query.shape[1]
            assert seqlen_Q_aligned % q_tile_size == 0
            assert logsumexp.shape[1] == seqlen_Q_aligned
            logger.debug(
                f"Padded q, dQ, o, dO, and lse with shape {old_shape} to {query.shape}."
            )

        d_query = torch.empty_like(query)
        d_key = torch.empty_like(key)
        d_value = torch.empty_like(value)

        if torch.are_deterministic_algorithms_enabled():
            raise RuntimeError(
                "Blackwell FMHA backward pass does not have a deterministic mode, "
                "but PyTorch's deterministic algorithms were enabled. To proceed, "
                "you must either disable torch's deterministic mode, or choose a "
                "different backend."
            )

        blackwell_fmha_backward(
            d_query,
            d_key,
            d_value,
            query,
            key,
            value,
            output,
            d_output,
            logsumexp,
            ctx.scale,
            q_tile_size,
            k_tile_size,
            seqlen_Q,
        )

        if seqlen_Q % q_tile_size != 0:
            d_query = d_query[:, :seqlen_Q, :, :]
            assert d_query.shape[1] == seqlen_Q

        return d_query, d_key, d_value, None, None, None, None


def cutlass_blackwell_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    scale: Optional[float] = None,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    run_persistent_kernel: bool = False,
    backward_q_tile_size: Optional[int] = None,
    backward_kv_tile_size: Optional[int] = None,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    fmha_tensor_checks(query, key, value, must_match_head_dims=True)

    assert can_run_cutlass_blackwell_fmha(query, key, value, raise_error=True)

    forward_config = check_cutlass_blackwell_fmha_forward_config(
        input_tensor=query, q_tile_size=q_tile_size, kv_tile_size=kv_tile_size
    )
    backward_config = check_cutlass_blackwell_fmha_backward_config(
        input_tensor=query,
        q_tile_size=backward_q_tile_size,
        kv_tile_size=backward_kv_tile_size,
    )

    scale = scale or query.shape[-1] ** -0.5

    output, lse = CutlassBlackwellFmhaAutogradFn.apply(
        query,
        key,
        value,
        scale,
        forward_config,
        backward_config,
        run_persistent_kernel,
    )

    if return_lse:
        return output, lse

    return output



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################


from torch import Tensor
from torch.nn.attention import sdpa_kernel, SDPBackend

from torch.nn.functional import scaled_dot_product_attention


def sdpa(q: Tensor, k: Tensor, v: Tensor, backend: str) -> Tensor:
    backends = []

    if backend == "xformers":
        backends = [SDPBackend.EFFICIENT_ATTENTION]

    elif backend == "fav2":
        backends = [SDPBackend.FLASH_ATTENTION]

    elif backend == "cudnn":
        backends = [SDPBackend.CUDNN_ATTENTION]

    else:
        raise NotImplementedError(
            f"Unrecognized SDPA backend {backend}. Choices are "
            "xformers, fav2, cudnn."
        )

    with sdpa_kernel(backends=backends):
        return scaled_dot_product_attention(q, k, v)



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################


from typing import Optional, Tuple

import torch

from ..backends import (
    can_run_cutlass_blackwell_fmha,
    can_run_cutlass_blackwell_fna,
    can_run_cutlass_fmha,
    can_run_cutlass_fna,
    can_run_cutlass_hopper_fmha,
    can_run_cutlass_hopper_fna,
    can_run_flex_attention,
    choose_backend,
    choose_fmha_backend,
    get_bwd_configs_for_cutlass_blackwell_fmha,
    get_bwd_configs_for_cutlass_blackwell_fna,
    get_bwd_configs_for_cutlass_fmha,
    get_bwd_configs_for_cutlass_fna,
    get_bwd_configs_for_cutlass_hopper_fmha,
    get_bwd_configs_for_cutlass_hopper_fna,
    get_compatible_backends,
    get_compatible_fmha_backends,
    get_configs_for_cutlass_blackwell_fmha,
    get_configs_for_cutlass_blackwell_fna,
    get_configs_for_cutlass_fmha,
    get_configs_for_cutlass_fna,
    get_configs_for_cutlass_hopper_fmha,
    get_configs_for_cutlass_hopper_fna,
    get_configs_for_flex_fmha,
    get_configs_for_flex_fna,
)
from ..types import DimensionType, KernelSchedule
from ..utils import log

from . import print_table

from .pretty_printer import opt_progress_bar
from .problem import Problem
from .profiling import measure_natten_runtime

logger = log.get_logger(__name__)

# TODO: these are duplicated from profiler.py
NATTEN_BACKENDS = ["cutlass-fna", "blackwell-fna", "hopper-fna", "flex-fna"]
NATTEN_FMHA_BACKENDS = ["cutlass-fmha", "blackwell-fmha", "hopper-fmha", "flex-fmha"]


def dry_run_for_backend(
    problem: Problem,
    backend: Optional[str],
    fmha_backend: Optional[str],
    backprop: bool,
    torch_compile: bool,
    max_configs: int,
    should_print: bool = True,
) -> Tuple[bool, list, list]:
    if backend is None and fmha_backend is None:
        raise ValueError(
            "Either backend or fmha_backend must be specified. "
            "If you're not a developer, please open an issue."
        )

    if backend is None and not problem.is_self_attn:
        raise ValueError(
            "FNA backend must be specified for non-self-attention case. "
            "If you're not a developer, please open an issue."
        )

    if fmha_backend is None and problem.is_self_attn:
        raise ValueError(
            "FMHA backend must be specified for self-attention case. "
            "If you're not a developer, please open an issue."
        )

    if backend is not None and backend not in NATTEN_BACKENDS:
        raise ValueError("--dry-run is only supported for NATTEN backends.")

    is_fmha = problem.is_self_attn

    torch_device = "cuda" if torch.cuda.is_available() else "cpu"
    q, k, v = problem.make_qkv_tensors(
        device=torch_device, requires_grad=backprop, heads_last=False, flatten=is_fmha
    )

    fwd_configs, bwd_configs = None, None
    fwd_config_keys, bwd_config_keys = None, None
    selected_backend = ""

    if is_fmha:
        assert fmha_backend is not None
        assert fmha_backend in NATTEN_FMHA_BACKENDS
        selected_backend = fmha_backend

        if fmha_backend == "blackwell-fmha":
            assert can_run_cutlass_blackwell_fmha(q, k, v, raise_error=True)
            fwd_configs = get_configs_for_cutlass_blackwell_fmha(q, k, v)  # type: ignore[assignment]
            bwd_configs = get_bwd_configs_for_cutlass_blackwell_fmha(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = ("q_tile_size", "kv_tile_size")
            bwd_config_keys = (
                "backward_q_tile_size",
                "backward_kv_tile_size",
            )

        elif fmha_backend == "hopper-fmha":
            assert can_run_cutlass_hopper_fmha(q, k, v, raise_error=True)
            fwd_configs = get_configs_for_cutlass_hopper_fmha(q, k, v)  # type: ignore[assignment]
            bwd_configs = get_bwd_configs_for_cutlass_hopper_fmha(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = (("q_tile_size", "kv_tile_size"), "kernel_schedule")  # type: ignore[assignment]
            bwd_config_keys = (
                "backward_q_tile_shape",
                "backward_kv_tile_shape",
            )

        elif fmha_backend == "cutlass-fmha":
            assert can_run_cutlass_fmha(q, k, v, raise_error=True)
            fwd_configs = get_configs_for_cutlass_fmha(q, k, v)  # type: ignore[assignment]
            bwd_configs = get_bwd_configs_for_cutlass_fmha(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = ("q_tile_size", "kv_tile_size")
            bwd_config_keys = (
                "backward_q_tile_size",
                "backward_kv_tile_size",
            )

        elif fmha_backend == "flex-fmha":
            assert can_run_flex_attention(
                q, k, v, torch_compile=torch_compile, raise_error=True
            )
            fwd_configs = get_configs_for_flex_fmha(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = ("q_tile_size", "kv_tile_size")

        else:
            raise NotImplementedError()

    else:
        assert backend is not None
        assert backend in NATTEN_BACKENDS
        selected_backend = backend

        if backend == "blackwell-fna":
            assert can_run_cutlass_blackwell_fna(q, k, v, raise_error=True)
            fwd_configs = get_configs_for_cutlass_blackwell_fna(q, k, v)  # type: ignore[assignment]
            bwd_configs = get_bwd_configs_for_cutlass_blackwell_fna(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = ("q_tile_shape", "kv_tile_shape")  # type: ignore[assignment]
            bwd_config_keys = (
                "backward_q_tile_shape",
                "backward_kv_tile_shape",
            )

        elif backend == "hopper-fna":
            assert can_run_cutlass_hopper_fna(q, k, v, raise_error=True)
            fwd_configs = get_configs_for_cutlass_hopper_fna(q, k, v)  # type: ignore[assignment]
            bwd_configs = get_bwd_configs_for_cutlass_hopper_fna(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = (("q_tile_shape", "kv_tile_shape"), "kernel_schedule")  # type: ignore[assignment]
            bwd_config_keys = (
                "backward_q_tile_shape",
                "backward_kv_tile_shape",
            )

        elif backend == "cutlass-fna":
            assert can_run_cutlass_fna(q, k, v, raise_error=True)
            fwd_configs = get_configs_for_cutlass_fna(q, k, v)  # type: ignore[assignment]
            bwd_configs = get_bwd_configs_for_cutlass_fna(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = ("q_tile_shape", "kv_tile_shape")  # type: ignore[assignment]
            bwd_config_keys = (
                "backward_q_tile_shape",
                "backward_kv_tile_shape",
            )

        elif backend == "flex-fna":
            assert can_run_flex_attention(
                q, k, v, torch_compile=torch_compile, raise_error=True
            )
            fwd_configs = get_configs_for_flex_fna(q, k, v)  # type: ignore[assignment]
            fwd_config_keys = ("q_tile_shape", "kv_tile_shape")  # type: ignore[assignment]

        else:
            raise NotImplementedError()

    # TODO: refactor this
    def print_configs(title, configs, keys) -> list:
        filtered_configs = []
        for cfg in configs:
            assert len(cfg) == len(keys)
            cfg_annotated = {}
            for key, val in zip(keys, cfg):
                if isinstance(key, tuple):
                    assert isinstance(key, tuple)
                    for key_sub, val_sub in zip(key, val):
                        assert isinstance(key_sub, str)
                        cfg_annotated[key_sub] = val_sub
                else:
                    assert isinstance(key, str)
                    cfg_annotated[key] = val

            filtered_configs.append(cfg_annotated)

        assert len(filtered_configs) > 0

        headers = filtered_configs[0].keys()
        assert all(x.keys() == headers for x in filtered_configs)
        values = [[str(y) for y in x.values()] for x in filtered_configs]

        if max_configs > 0 and len(values) > max_configs:
            values = values[:max_configs]
            values.append(["..." for _ in headers])

        if should_print:
            print_table(title, headers, values, has_footer=False)

        return filtered_configs

    assert fwd_configs is not None and fwd_config_keys is not None
    if len(fwd_configs) < 1:
        return False, [], []

    fwd_configs_out = print_configs(
        title=f"Backend: {selected_backend}\nForward pass configurations",
        configs=fwd_configs,
        keys=fwd_config_keys,
    )

    bwd_configs_out = []
    if bwd_configs is not None:
        assert bwd_config_keys is not None
        bwd_configs_out = print_configs(
            title=f"Backend: {selected_backend}\nBackward pass configurations",
            configs=bwd_configs,
            keys=bwd_config_keys,
        )

    return True, fwd_configs_out, bwd_configs_out


def dry_run(
    problem: Problem,
    backend: Optional[str],
    fmha_backend: Optional[str],
    backprop: bool,
    torch_compile: bool,
    max_configs: int,
):
    if backend is not None and backend not in NATTEN_BACKENDS:
        raise ValueError("Dry run is only supported for NATTEN backends.")

    torch_device = "cuda" if torch.cuda.is_available() else "cpu"
    is_fmha = problem.is_self_attn

    backends = None
    fmha_backends = None
    if backend is None:
        q, k, v = problem.make_qkv_tensors(
            device=torch_device, requires_grad=backprop, heads_last=False, flatten=False
        )

        backends = get_compatible_backends(q, k, v, torch_compile=torch_compile)
    else:
        backends = [backend]

    if fmha_backend is None:
        q_flat, k_flat, v_flat = problem.make_qkv_tensors(
            device=torch_device, requires_grad=backprop, heads_last=False, flatten=True
        )

        fmha_backends = get_compatible_fmha_backends(
            q_flat, k_flat, v_flat, torch_compile=torch_compile
        )
    else:
        fmha_backends = [fmha_backend]

    compatible_backends = fmha_backends if is_fmha else backends

    ran_successfully = False
    for b in compatible_backends:
        print(f"Use case is compatible with backend {b}.")
        result, _, _ = dry_run_for_backend(
            problem=problem,
            backend=None if is_fmha else b,
            fmha_backend=b if is_fmha else None,
            backprop=backprop,
            torch_compile=torch_compile,
            max_configs=max_configs,
        )
        ran_successfully = ran_successfully or result

    if not ran_successfully:
        raise RuntimeError(
            "No backends are compatible with this use case, or you selected an incompatible backend. "
            "Run with NATTEN_LOG_LEVEL=DEBUG to find out why."
        )


def find_configs(
    problem: Problem,
    backend: Optional[str],
    fmha_backend: Optional[str],
    backprop: bool,
    torch_compile: bool,
) -> Tuple[str, str, list, list]:
    torch_device = "cuda" if torch.cuda.is_available() else "cpu"
    q, k, v = problem.make_qkv_tensors(
        device=torch_device, requires_grad=backprop, heads_last=False, flatten=False
    )
    q_flat, k_flat, v_flat = problem.make_qkv_tensors(
        device=torch_device, requires_grad=backprop, heads_last=False, flatten=True
    )

    backend = backend or choose_backend(q, k, v, torch_compile=torch_compile)
    fmha_backend = fmha_backend or choose_fmha_backend(
        q_flat, k_flat, v_flat, torch_compile=torch_compile
    )

    result, fwd_configs, bwd_configs = dry_run_for_backend(
        problem=problem,
        backend=backend,
        fmha_backend=fmha_backend,
        backprop=backprop,
        torch_compile=torch_compile,
        max_configs=0,
        should_print=False,
    )

    if not result:
        raise RuntimeError(
            f"Selected backends ({backend=}, {fmha_backend=}) are not compatible "
            "with your use case."
        )

    return backend, fmha_backend, fwd_configs, bwd_configs


def optimize(
    problem: Problem,
    warmup_steps: int,
    backend: str,
    fmha_backend: str,
    backprop: bool,
    q_tile: Optional[DimensionType],
    kv_tile: Optional[DimensionType],
    backward_q_tile: Optional[DimensionType],
    backward_kv_tile: Optional[DimensionType],
    persistent: bool,
    schedule: Optional[str],
    torch_compile: bool,
    q_tile_size: Optional[int],
    kv_tile_size: Optional[int],
    backward_q_tile_size: Optional[int],
    backward_kv_tile_size: Optional[int],
) -> Tuple[
    str,
    str,
    Optional[DimensionType],
    Optional[DimensionType],
    Optional[DimensionType],
    Optional[DimensionType],
    Optional[int],
    Optional[int],
    Optional[int],
    Optional[int],
    Optional[KernelSchedule],
]:
    if backend is not None and backend not in NATTEN_BACKENDS:
        raise ValueError(
            f"Optimize is only available for NATTEN backends, got {backend=}."
        )

    opt_backend, opt_fmha_backend, fwd_configs, bwd_configs = find_configs(
        problem,
        backend=backend,
        fmha_backend=fmha_backend,
        backprop=backprop,
        torch_compile=torch_compile,
    )

    # Override with user inputs, if any
    if q_tile is not None:
        assert kv_tile is not None
        fwd_configs = (
            [
                {
                    "q_tile_shape": q_tile,
                    "kv_tile_shape": kv_tile,
                    "q_tile_size": q_tile_size,
                    "kv_tile_size": kv_tile_size,
                    "kernel_schedule": schedule,
                }
            ]
            if opt_backend == "hopper-fna"
            else [
                {
                    "q_tile_shape": q_tile,
                    "kv_tile_shape": kv_tile,
                    "q_tile_size": q_tile_size,
                    "kv_tile_size": kv_tile_size,
                }
            ]
        )

    if backward_q_tile is not None and backprop:
        assert backward_kv_tile is not None
        # assert opt_backend == "cutlass-fna"
        bwd_configs = [
            {
                "backward_q_tile_shape": backward_q_tile,
                "backward_kv_tile_shape": backward_kv_tile,
                "backward_q_tile_size": backward_q_tile_size,
                "backward_kv_tile_size": backward_kv_tile_size,
            }
        ]

    def run_optimize(configs, run_backprop: bool):
        best_time_str = None
        best_time = None
        best_config = None

        for i, cfg in opt_progress_bar(enumerate(configs), total=len(configs)):
            logger.debug(f"Testing config [{i+1}/{len(configs)}]")
            # Manually measure time using cuda events instead of profiling due to an
            # intermittent bug.

            runtime = measure_natten_runtime(
                problem=problem,
                backend=opt_backend,
                fmha_backend=opt_fmha_backend,
                run_persistent_kernel=persistent,
                torch_compile=torch_compile,
                warmup_steps=warmup_steps,
                disable_backward=not run_backprop,
                **cfg,
            )
            runtime_str = f"{runtime:.2f} ms"

            # logged_ops = do_profile(
            #     problem=problem,
            #     backend=opt_backend,
            #     fmha_backend=opt_fmha_backend,
            #     persistent=persistent,
            #     torch_compile=torch_compile,
            #     warmup_steps=warmup_steps,
            #     backprop=run_backprop,
            #     max_retries=20,
            #     fail_on_capture_error=False, # just skip the config if it doesn't get captured.
            #     **cfg,
            # )
            # if logged_ops is None or len(logged_ops) < 1:
            #     logger.debug(f"Could not profile configuration due to profiler bug, skipping...")
            #     continue
            # total = sum(logged_ops)
            # runtime = total.time
            # runtime_str = total.time_str

            logger.debug(f"Configuration: {cfg=}, runtime: {runtime_str}")
            if best_time is None or runtime < best_time:
                best_time_str = runtime_str
                best_time = runtime
                best_config = cfg

        assert best_time is not None
        assert best_config is not None

        return best_config, best_time_str

    best_cfg = {
        "backend": opt_backend,
        "fmha_backend": opt_fmha_backend,
    }

    print()
    print(f"Searching {len(fwd_configs)} forward pass configs")
    best_fwd_cfg, best_fwd_time = run_optimize(fwd_configs, run_backprop=False)
    logger.debug(
        f"Best forward configuration: {best_fwd_cfg} (runtime: {best_fwd_time})"
    )
    best_cfg.update(best_fwd_cfg)

    if backprop and len(bwd_configs) > 0:
        print()
        print(f"Searching {len(bwd_configs)} backward pass configs")
        best_bwd_cfg, best_bwd_time = run_optimize(bwd_configs, run_backprop=True)
        logger.debug(
            f"Best backward configuration: {best_bwd_cfg} (runtime: {best_bwd_time})"
        )
        best_cfg.update(best_bwd_cfg)

    print()
    print_table(
        "Best configuration",
        ["Parameter", "Value"],
        [[k, str(v)] for k, v in best_cfg.items()],
        has_footer=False,
    )
    print()

    assert ("q_tile_size" in best_cfg and "kv_tile_size" in best_cfg) or (
        "q_tile_shape" in best_cfg and "kv_tile_shape" in best_cfg
    )
    assert not ("backward_q_tile_size" in best_cfg) ^ (
        "backward_kv_tile_size" in best_cfg
    )
    assert not ("backward_q_tile_shape" in best_cfg) ^ (
        "backward_kv_tile_shape" in best_cfg
    )

    opt_q_tile: Optional[DimensionType] = None
    opt_kv_tile: Optional[DimensionType] = None
    opt_backward_q_tile: Optional[DimensionType] = None
    opt_backward_kv_tile: Optional[DimensionType] = None
    opt_q_tile_size: Optional[int] = None
    opt_kv_tile_size: Optional[int] = None
    opt_backward_q_tile_size: Optional[int] = None
    opt_backward_kv_tile_size: Optional[int] = None

    if "q_tile_shape" in best_cfg and "kv_tile_shape" in best_cfg:
        opt_q_tile = best_cfg["q_tile_shape"]  # type: ignore[assignment]
        opt_kv_tile = best_cfg["kv_tile_shape"]  # type: ignore[assignment]

    if "backward_q_tile_shape" in best_cfg and "backward_kv_tile_shape" in best_cfg:
        opt_backward_q_tile = best_cfg["backward_q_tile_shape"]  # type: ignore[assignment]
        opt_backward_kv_tile = best_cfg["backward_kv_tile_shape"]  # type: ignore[assignment]

    if "q_tile_size" in best_cfg and "kv_tile_size" in best_cfg:
        opt_q_tile_size = best_cfg["q_tile_size"]  # type: ignore[assignment]
        opt_kv_tile_size = best_cfg["kv_tile_size"]  # type: ignore[assignment]

    if "backward_q_tile_size" in best_cfg and "backward_kv_tile_size" in best_cfg:
        opt_backward_q_tile_size = best_cfg["backward_q_tile_size"]  # type: ignore[assignment]
        opt_backward_kv_tile_size = best_cfg["backward_kv_tile_size"]  # type: ignore[assignment]

    opt_kernel_schedule: Optional[KernelSchedule] = (
        None if "kernel_schedule" not in best_cfg else best_cfg["kernel_schedule"]  # type: ignore[assignment]
    )

    return (
        opt_backend,
        opt_fmha_backend,
        opt_q_tile,
        opt_kv_tile,
        opt_backward_q_tile,
        opt_backward_kv_tile,
        opt_q_tile_size,
        opt_kv_tile_size,
        opt_backward_q_tile_size,
        opt_backward_kv_tile_size,
        opt_kernel_schedule,
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

import functools
from collections.abc import Sequence
from typing import Any, Optional, Tuple

import torch  # noqa: F401
from torch import Tensor

from ..types import CausalArgType, DimensionType, KernelSchedule
from ..utils.tuples import create_causal_arg_from_bool, create_dim_from_int

from . import log

logger = log.get_logger(__name__)


def log_or_raise_error(
    msg: str, raise_error: bool = False, exception: Any = RuntimeError
):
    if raise_error:
        raise exception(msg)
    else:
        logger.debug(msg)


def _universal_tensor_checks(
    query: Tensor, key: Tensor, value: Tensor, raise_error: bool = True
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if query.is_sparse or key.is_sparse or value.is_sparse:
        target_fn(
            "NATTEN does not support sparse tensors.", exception=NotImplementedError
        )
        return False

    if query.is_nested or key.is_nested or value.is_nested:
        target_fn(
            "NATTEN does not support nested tensors.", exception=NotImplementedError
        )
        return False

    if query.device != key.device or query.device != value.device:
        target_fn(
            "Query, key, and value must be on the same device, "
            f"got {query.device=}, {key.device=}, {value.device=}.",
            exception=ValueError,
        )
        return False

    if query.dtype != key.dtype or query.dtype != value.dtype:
        target_fn(
            "Query, key, and value must assume the same data type, "
            f"got {query.dtype=}, {key.dtype=}, {value.dtype=}.",
            exception=ValueError,
        )
        return False

    return True


def na_tensor_checks(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    must_match_head_dims: bool = False,
    raise_error: bool = True,
    backend_name: Optional[str] = None,
) -> bool:
    backend_name = backend_name or "This operation/backend"
    if not _universal_tensor_checks(query, key, value):
        return False

    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if query.dim() != key.dim() or query.dim() != value.dim():
        target_fn(
            "Query, key, and value must have the same rank, "
            f"got {query.dim()=}, {key.dim()=}, {value.dim()=}.",
            exception=ValueError,
        )
        return False

    if query.dim() not in [4, 5, 6]:
        target_fn(
            "Expected 4-D, 5-D, or 6-D tensors as inputs (corresponding to NA1D, NA2D, and NA3D), "
            f"got {query.dim()=}.",
            exception=ValueError,
        )
        return False

    if query.shape != key.shape or query.shape[:-1] != value.shape[:-1]:
        target_fn(
            "Neighborhood attention ops expect Q, K, and V to have the same shape (except value "
            f"head dim), got {query.shape=}, {key.shape=}, {value.shape=}.",
            exception=ValueError,
        )
        return False

    if must_match_head_dims and query.shape[-1] != value.shape[-1]:
        target_fn(
            f"{backend_name} does not support different head dims for QK and V, got "
            f"{query.shape[-1]=}, {value.shape[-1]=}.",
            exception=ValueError,
        )
        return False

    return True


def fmha_tensor_checks(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    must_match_head_dims: bool = False,
    raise_error: bool = True,
    backend_name: Optional[str] = None,
) -> bool:
    backend_name = backend_name or "This operation/backend"
    if not _universal_tensor_checks(query, key, value):
        return False

    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if query.dim() != key.dim() or query.dim() != value.dim():
        target_fn(
            "Query, key, and value must have the same rank, "
            f"got {query.dim()=}, {key.dim()=}, {value.dim()=}.",
            exception=ValueError,
        )
        return False

    if query.dim() != 4:
        target_fn(
            "Expected 4-D tensors as inputs to FMHA, " f"got {query.dim()=}.",
            exception=ValueError,
        )
        return False

    if query.shape[-1] != key.shape[-1]:
        target_fn(
            f"Q and K head dims must match, got {query.shape[-1]=}, {key.shape[-1]=}.",
            exception=ValueError,
        )
        return False

    if must_match_head_dims and query.shape[-1] != value.shape[-1]:
        target_fn(
            f"{backend_name} does not support different head dims for QK and V, got "
            f"{query.shape[-1]=}, {value.shape[-1]=}.",
            exception=ValueError,
        )
        return False

    if query.shape[0] != key.shape[0] or query.shape[0] != value.shape[0]:
        target_fn(
            "Q, K, and V must match in batch size, got "
            f"{query.shape[0]=}, {key.shape[0]=}, {value.shape[0]=}.",
            exception=ValueError,
        )
        return False

    if key.shape[1] != value.shape[1]:
        target_fn(
            f"K and V must match in sequence length, got {key.shape[1]=}, {value.shape[1]=}.",
            exception=ValueError,
        )
        return False

    if query.shape[2] != key.shape[2] or query.shape[2] != value.shape[2]:
        target_fn(
            "NATTEN operations do not support GQA/MQA, therefore number of heads in Q, K, and V "
            f"must match, got {query.shape[2]=}, {key.shape[2]=}, {value.shape[2]=}.",
            exception=ValueError,
        )
        return False

    return True


def additional_kv_tensor_checks(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    add_key: Optional[Tensor] = None,
    add_value: Optional[Tensor] = None,
    must_match_head_dims: bool = False,
):

    if (add_key is not None) ^ (add_value is not None):
        raise ValueError(
            "`additional_keys` and `additional_values` must be either both Tensors or None."
        )

    if add_key is None:
        return

    assert add_key is not None and add_value is not None

    _universal_tensor_checks(query, add_key, add_value)

    if query.shape[-1] != add_key.shape[-1]:
        raise ValueError(
            f"Q and K head dims must match, got {query.shape[-1]=}, {add_key.shape[-1]=}."
        )

    if must_match_head_dims and query.shape[-1] != add_value.shape[-1]:
        raise ValueError(
            "This operation does not support different head dims for QK and V, got "
            f"{query.shape[-1]=}, {add_value.shape[-1]=}."
        )

    if query.shape[0] != add_key.shape[0] or query.shape[0] != add_value.shape[0]:
        raise ValueError(
            "Q, additional K, and additional V must match in batch size, got "
            f"{query.shape[0]=}, {add_key.shape[0]=}, {add_value.shape[0]=}."
        )

    if add_key.shape[1] != add_value.shape[1]:
        raise ValueError(
            f"Additional K and V must match in sequence length, got {add_key.shape[1]=}, "
            f"{add_value.shape[1]=}."
        )

    if query.shape[-2] != add_key.shape[-2] or query.shape[-2] != add_value.shape[-2]:
        raise ValueError(
            "NATTEN operations do not support GQA/MQA, therefore number of heads in Q, K, and V "
            f"must match, got {query.shape[-2]=}, {add_key.shape[-2]=}, {add_value.shape[-2]=}."
        )

    if key.shape[0] != add_key.shape[0] or value.shape[0] != add_value.shape[0]:
        raise ValueError(
            "Additional key/value tokens must match the self attention key/value tokens in batch "
            f"size, got {key.shape[0]=} != {add_key.shape[0]=}, and "
            f"{value.shape[0]=} != {add_value.shape[0]=}."
        )

    if key.shape[-2] != add_key.shape[-2] or value.shape[-2] != add_value.shape[-2]:
        raise ValueError(
            "Additional key/value tokens must match the self attention key/value tokens in number "
            f"of heads, got {key.shape[-2]=} != {add_key.shape[-2]=}, and "
            f"{value.shape[-2]=} != {add_value.shape[-2]=}."
        )

    if key.shape[-1] != add_key.shape[-1] or value.shape[-1] != add_value.shape[-1]:
        raise ValueError(
            "Additional key/value tokens must match the self attention key/value tokens in head "
            f"dim, got {key.shape[-1]=} != {add_key.shape[-1]=}, and "
            f"{value.shape[-1]=} != {add_value.shape[-1]=}."
        )


def check_input_size_arg(na_dim: int, input_size: Any) -> DimensionType:
    assert na_dim > 0 and na_dim < 4
    if (
        isinstance(input_size, Sequence)
        and len(input_size) == na_dim
        and all(isinstance(x, int) and x > 1 for x in input_size)
    ):
        return tuple(x for x in input_size)

    if isinstance(input_size, int) and input_size > 1:
        return create_dim_from_int(na_dim, value=input_size)

    raise ValueError(
        "Invalid value for `input_size`; expected an integer or iterable of integers, all >= 2, "
        f"got {type(input_size)}"
    )


def check_kernel_size_arg(na_dim: int, kernel_size: Any) -> DimensionType:
    assert na_dim > 0 and na_dim < 4
    if (
        isinstance(kernel_size, Sequence)
        and len(kernel_size) == na_dim
        and all(isinstance(x, int) and x > 1 for x in kernel_size)
    ):
        return tuple(x for x in kernel_size)

    if isinstance(kernel_size, int) and kernel_size > 1:
        return create_dim_from_int(na_dim, value=kernel_size)

    raise ValueError(
        "Invalid value for `kernel_size`; expected an integer or iterable of integers, all >= 2, "
        f"got {type(kernel_size)}"
    )


def check_stride_arg(na_dim: int, stride: Any) -> DimensionType:
    assert na_dim > 0 and na_dim < 4
    if stride is None:
        return create_dim_from_int(na_dim, value=1)

    if (
        isinstance(stride, Sequence)
        and len(stride) == na_dim
        and all(isinstance(x, int) and x > 0 for x in stride)
    ):
        return tuple(x for x in stride)

    if isinstance(stride, int) and stride > 0:
        return create_dim_from_int(na_dim, value=stride)

    raise ValueError(
        "Invalid value for `stride`; expected an integer or tuple of positive integers, "
        f"got {type(stride)}"
    )


def check_dilation_arg(na_dim: int, dilation: Any) -> DimensionType:
    assert na_dim > 0 and na_dim < 4
    if dilation is None:
        return create_dim_from_int(na_dim, value=1)

    if (
        isinstance(dilation, Sequence)
        and len(dilation) == na_dim
        and all(isinstance(x, int) and x > 0 for x in dilation)
    ):
        return tuple(x for x in dilation)

    if isinstance(dilation, int) and dilation > 0:
        return create_dim_from_int(na_dim, value=dilation)

    raise ValueError(
        "Invalid value for `dilation`; expected an integer or tuple of positive integers, "
        f"got {type(dilation)}"
    )


def check_causal_arg(na_dim: int, is_causal: Any) -> CausalArgType:
    assert na_dim > 0 and na_dim < 4

    if is_causal is None:
        return create_causal_arg_from_bool(na_dim, value=False)

    if (
        isinstance(is_causal, Sequence)
        and len(is_causal) == na_dim
        and all(isinstance(c, bool) for c in is_causal)
    ):
        return tuple(c for c in is_causal)

    if isinstance(is_causal, bool):
        return create_causal_arg_from_bool(na_dim, value=is_causal)

    raise ValueError(
        "Invalid value for `is_causal`; expected a boolean or tuple of booleans, "
        f"got {type(is_causal)}"
    )


def check_all_args(
    na_dim: int, kernel_size: Any, stride: Any, dilation: Any, is_causal: Any
) -> Tuple[DimensionType, DimensionType, DimensionType, CausalArgType]:
    kernel_size_out, stride_out, dilation_out, is_causal_out = (
        check_kernel_size_arg(na_dim, kernel_size),
        check_stride_arg(na_dim, stride),
        check_dilation_arg(na_dim, dilation),
        check_causal_arg(na_dim, is_causal),
    )

    return kernel_size_out, stride_out, dilation_out, is_causal_out


def check_args_against_input(
    input_tensor: Tensor,
    kernel_size: DimensionType,
    stride: DimensionType,
    dilation: DimensionType,
    is_causal: CausalArgType,
):
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3
    input_size = input_tensor.shape[1 : 1 + na_dim]

    if any(k * d > x for x, k, d in zip(input_size, kernel_size, dilation)):
        raise ValueError(
            "The product of kernel size and dilation cannot be larger than input size "
            f"along any dimension, got {input_size=} ({input_tensor.shape=}), "
            f"{kernel_size=}, {dilation=}."
        )

    if any(s > k for k, s in zip(kernel_size, stride)):
        raise ValueError(
            "Stride cannot be larger than kernel size along any dimension, got "
            f"{kernel_size=}, {stride=}."
        )


def is_self_attention(
    input_tensor: Tensor, kernel_size: DimensionType, is_causal: CausalArgType
):
    assert input_tensor.dim() in [4, 5, 6]
    na_dim = input_tensor.dim() - 3
    input_size = input_tensor.shape[1 : 1 + na_dim]

    return all(k == x and not c for x, k, c in zip(input_size, kernel_size, is_causal))


def check_tile_shape(
    tile_shape: Any,
) -> DimensionType:
    if (
        isinstance(tile_shape, Sequence)
        and len(tile_shape) <= 3
        and all(isinstance(x, int) for x in tile_shape)
    ):
        return tuple(x for x in tile_shape)

    raise ValueError(
        f"Unsupported value for tile shape; expected an iterable of at most 3 integers, "
        f"got {type(tile_shape)}: {tile_shape}"
    )


def check_kernel_schedule(kernel_schedule: Any) -> Optional[KernelSchedule]:
    if kernel_schedule is None:
        return None

    if isinstance(kernel_schedule, KernelSchedule):
        return kernel_schedule

    if kernel_schedule == "non":
        return KernelSchedule.NonPersistent
    elif kernel_schedule == "coop":
        return KernelSchedule.WarpSpecializedCooperative
    elif kernel_schedule == "pp":
        return KernelSchedule.WarpSpecializedPingpong

    raise ValueError(
        f"Kernel schedule {kernel_schedule} is invalid; choices are: "
        "`non` (non-persistent), `coop` (warp-specialized cooperative), and "
        "`pp` (warp-specialized ping-ponging)."
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import Optional

import torch
from torch import Size, Tensor


def _get_expected_attn_shape(input_tensor: Tensor, attention_dim: int) -> Size:
    shape = [x for x in input_tensor.shape[:-1]] + [attention_dim]
    return Size(shape)


def make_attn_tensor_from_input(input_tensor: Tensor, attention_dim: int) -> Tensor:
    return torch.empty(
        _get_expected_attn_shape(input_tensor, attention_dim),
        device=input_tensor.device,
        dtype=input_tensor.dtype,
        requires_grad=input_tensor.requires_grad,
    )


def check_additional_keys(
    input_tensor: Tensor, additional_keys: Optional[Tensor]
) -> int:
    if additional_keys is None:
        return 0

    if additional_keys.dim() != 4:
        raise ValueError(
            "Additional tokens have to be shaped as a rank-4 tensor; "
            f"got {additional_keys.dim()}."
        )
    batch_size, heads, tokens, dim = additional_keys.shape
    expected_batch_size = input_tensor.shape[0]
    expected_heads = input_tensor.shape[1]
    expected_dim = input_tensor.shape[-1]
    if (
        batch_size != expected_batch_size
        or expected_heads != heads
        or expected_dim != dim
    ):
        raise ValueError(
            "Shape mismatch between input tensor and additional tokens; "
            "they must match in batch size, heads, and dim per head. "
            f"Got {input_tensor.shape=}, {additional_keys.shape=}."
        )
    return tokens


def check_additional_values(
    attn_tensor: Tensor,
    additional_values: Optional[Tensor],
    value: Tensor,
    expected_attn_weights: int,
) -> int:
    if additional_values is None and attn_tensor.shape[-1] == expected_attn_weights:
        return 0
    if additional_values is None:
        raise ValueError(
            f"Expected {expected_attn_weights} attention weights per token, "
            f"got {attn_tensor.shape[-1]=}."
        )

    if additional_values.dim() != 4:
        raise ValueError(
            "Additional tokens have to be shaped as a rank-4 tensor; "
            f"got {additional_values.dim()}."
        )

    if additional_values.shape[-1] != value.shape[-1]:
        raise ValueError(
            "Additional value tokens must match the dimension of the "
            f"rest of the tokens, got {additional_values.shape[-1]=} != "
            f"{value.shape[-1]=}."
        )

    batch_size, heads, tokens, dim = additional_values.shape
    if tokens + expected_attn_weights != attn_tensor.shape[-1]:
        raise ValueError(
            f"Expected {expected_attn_weights + tokens} attention weights per token, "
            f"got {attn_tensor.shape[-1]=}."
        )
    expected_batch_size = attn_tensor.shape[0]
    expected_heads = attn_tensor.shape[1]
    if batch_size != expected_batch_size or expected_heads != heads:
        raise ValueError(
            "Shape mismatch between attention tensor and additional tokens; "
            "they must match in batch size and heads. "
            f"Got {attn_tensor.shape=}, {additional_values.shape=}."
        )
    return tokens



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
import functools
import math

import torch
from torch import Tensor

from ..._environment import _IS_TORCH_COMPILE_SUPPORTED, _TORCH_VERSION
from ..._libnatten import HAS_LIBNATTEN
from ...context import is_flex_compile_allowed, is_flex_compile_backprop_allowed
from ...utils.checks import fmha_tensor_checks, log_or_raise_error, na_tensor_checks
from ...utils.device import get_device_cc, is_cpu, is_cuda, is_rocm


### Blackwell FMHA/FNA


def can_run_cutlass_blackwell_fmha(
    query: Tensor, key: Tensor, value: Tensor, raise_error: bool = False
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if not HAS_LIBNATTEN:
        target_fn("Can't run Blackwell FMHA; NATTEN was not built with libnatten.")
        return False

    if not fmha_tensor_checks(
        query,
        key,
        value,
        must_match_head_dims=True,
        raise_error=raise_error,
        backend_name="Blackwell FMHA",
    ):
        return False

    if query.dim() != 4:
        target_fn(
            f"Blackwell FMHA expects rank-4 input tensors, got {query.shape=}.",
            exception=ValueError,
        )
        return False

    if not is_cuda(query.device):
        target_fn("Can't run Blackwell FMHA; not a CUDA tensor.")
        return False

    device_cc = get_device_cc(query.device)

    if device_cc != 100:
        target_fn(
            "Can't run Blackwell FMHA; tensor was on CUDA device with "
            f"compute capability {device_cc}, expected 100."
        )
        return False

    if query.requires_grad and torch.are_deterministic_algorithms_enabled():
        target_fn(
            "Can't run Blackwell FMHA; its backprop does not have a deterministic mode, but "
            "PyTorch's deterministic mode was enabled.",
            exception=NotImplementedError,
        )
        return False

    head_dim = query.shape[-1]
    head_dim_v = value.shape[-1]

    if head_dim != head_dim_v:
        target_fn(
            "Can't run Blackwell FMHA; it does not support different head dims for QK and V, "
            f"got {head_dim=}, {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    if query.dtype not in [torch.float16, torch.bfloat16]:
        target_fn(
            "Can't run Blackwell FMHA; it only supports FP16 and BF16 for now.",
            exception=ValueError,
        )
        return False

    if head_dim not in [32, 64, 128]:
        target_fn(
            "Can't run Blackwell FMHA; it only supports head dims 32, 64, and 128 for now.",
            exception=NotImplementedError,
        )
        return False

    return True


def can_run_cutlass_blackwell_fna(
    query: Tensor, key: Tensor, value: Tensor, raise_error: bool = False
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if not HAS_LIBNATTEN:
        target_fn("Can't run Blackwell FNA; NATTEN was not built with libnatten.")
        return False

    if not na_tensor_checks(
        query,
        key,
        value,
        must_match_head_dims=True,
        raise_error=raise_error,
        backend_name="Blackwell FNA",
    ):
        return False

    if query.dim() not in [4, 5, 6]:
        target_fn(
            "Blackwell FNA expects 4-D, 5-D, or 6-D tensors as inputs (corresponding to NA1D, NA2D, and NA3D), "
            f"got {query.dim()=}.",
            exception=ValueError,
        )
        return False

    if not is_cuda(query.device):
        target_fn("Can't run Blackwell FNA; not a CUDA tensor.")
        return False

    device_cc = get_device_cc(query.device)

    if device_cc != 100:
        target_fn(
            "Can't run Blackwell FNA; tensor was on CUDA device with "
            f"compute capability {device_cc}, expected 100."
        )
        return False

    if query.requires_grad and torch.are_deterministic_algorithms_enabled():
        target_fn(
            "Can't run Blackwell FMHA; its backprop does not have a deterministic mode, but "
            "PyTorch's deterministic mode was enabled.",
            exception=NotImplementedError,
        )
        return False

    head_dim = query.shape[-1]
    head_dim_v = value.shape[-1]

    if head_dim != head_dim_v:
        target_fn(
            "Can't run Blackwell FNA; it does not support different head dims for QK and V, "
            f"got {head_dim=}, {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    if query.dtype not in [torch.float16, torch.bfloat16]:
        target_fn(
            "Can't run Blackwell FNA; it only supports FP16 and BF16 for now.",
            exception=ValueError,
        )
        return False

    if head_dim not in [32, 64, 128]:
        target_fn(
            "Can't run Blackwell FNA; it only supports head dims 32, 64, and 128 for now.",
            exception=NotImplementedError,
        )
        return False

    return True


### Hopper FMHA/FNA


def can_run_cutlass_hopper_fmha(
    query: Tensor, key: Tensor, value: Tensor, raise_error: bool = False
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if not HAS_LIBNATTEN:
        target_fn("Can't run Hopper FMHA; NATTEN was not built with libnatten.")
        return False

    if not fmha_tensor_checks(
        query,
        key,
        value,
        must_match_head_dims=True,
        raise_error=raise_error,
        backend_name="Hopper FMHA",
    ):
        return False

    if query.dim() != 4:
        target_fn(
            f"Hopper FMHA expects rank-4 input tensors, got {query.shape=}.",
            exception=ValueError,
        )
        return False

    if not is_cuda(query.device):
        target_fn("Can't run Hopper FMHA; not a CUDA tensor.")
        return False

    device_cc = get_device_cc(query.device)

    if device_cc != 90:
        target_fn(
            "Can't run Hopper FMHA; tensor was on CUDA device with "
            f"compute capability {device_cc}, expected 90."
        )
        return False

    head_dim = query.shape[-1]
    head_dim_v = value.shape[-1]

    if head_dim != head_dim_v:
        target_fn(
            "Can't run Hopper FMHA; it does not support different head dims for QK and V, "
            f"got {head_dim=}, {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    if query.requires_grad and head_dim not in [32, 64, 128]:
        target_fn(
            f"Can't run Hopper FMHA; it does not support backpropagation for {head_dim=} yet; "
            "only head dims 32, 64, and 128 are allowed.",
            exception=NotImplementedError,
        )
        return False

    if query.requires_grad and torch.are_deterministic_algorithms_enabled():
        target_fn(
            "Can't run Hopper FMHA; its backprop does not have a deterministic mode, but "
            "PyTorch's deterministic mode was enabled.",
            exception=NotImplementedError,
        )
        return False

    if query.dtype not in [torch.float16, torch.bfloat16]:
        target_fn(
            "Can't run Hopper FMHA; it only supports FP16 and BF16 for now.",
            exception=ValueError,
        )
        return False

    if head_dim not in [32, 64, 128, 256]:
        target_fn(
            "Can't run Hopper FMHA; it only supports head dims 32, 64, 128, and 256 for now.",
            exception=NotImplementedError,
        )
        return False

    return True


def can_run_cutlass_hopper_fna(
    query: Tensor, key: Tensor, value: Tensor, raise_error: bool = False
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if not HAS_LIBNATTEN:
        target_fn("Can't run Hopper FNA; NATTEN was not built with libnatten.")
        return False

    if not na_tensor_checks(
        query,
        key,
        value,
        must_match_head_dims=True,
        raise_error=raise_error,
        backend_name="Hopper FNA",
    ):
        return False

    if query.dim() not in [4, 5, 6]:
        target_fn(
            "Hopper FNA expects 4-D, 5-D, or 6-D tensors as inputs (corresponding to NA1D, NA2D, and NA3D), "
            f"got {query.dim()=}.",
            exception=ValueError,
        )
        return False

    if not is_cuda(query.device):
        target_fn("Can't run Hopper FNA; not a CUDA tensor.")
        return False

    device_cc = get_device_cc(query.device)

    if device_cc != 90:
        target_fn(
            "Can't run Hopper FNA; tensor was on CUDA device with "
            f"compute capability {device_cc}, expected 90."
        )
        return False

    head_dim = query.shape[-1]
    head_dim_v = value.shape[-1]

    if head_dim != head_dim_v:
        target_fn(
            "Can't run Hopper FNA; it does not support different head dims for QK and V, "
            f"got {head_dim=}, {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    if query.requires_grad and head_dim not in [32, 64, 128]:
        target_fn(
            f"Can't run Hopper FNA; it does not support backpropagation for {head_dim=} yet; "
            "only head dims 32, 64, and 128 are allowed.",
            exception=NotImplementedError,
        )
        return False

    if query.requires_grad and torch.are_deterministic_algorithms_enabled():
        target_fn(
            "Can't run Hopper FNA; its backprop does not have a deterministic mode, but "
            "PyTorch's deterministic mode was enabled.",
            exception=NotImplementedError,
        )
        return False

    if query.dtype not in [torch.float16, torch.bfloat16]:
        target_fn(
            "Can't run Hopper FNA; it only supports FP16 and BF16 for now.",
            exception=ValueError,
        )
        return False

    if head_dim not in [32, 64, 128, 256]:
        target_fn(
            "Can't run Hopper FNA; it only supports head dims 32, 64, 128, and 256 for now.",
            exception=NotImplementedError,
        )
        return False

    return True


### CUTLASS FMHA/FNA


def can_run_cutlass_fmha(
    query: Tensor, key: Tensor, value: Tensor, raise_error: bool = False
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if not HAS_LIBNATTEN:
        target_fn("Can't run CUTLASS FMHA; NATTEN was not built with libnatten.")
        return False

    if not fmha_tensor_checks(
        query,
        key,
        value,
        must_match_head_dims=False,
        raise_error=raise_error,
        backend_name="CUTLASS FMHA",
    ):
        return False

    if query.dim() != 4:
        target_fn(
            f"FMHA expects rank-4 input tensors, got {query.shape=}.",
            exception=ValueError,
        )
        return False

    if not is_cuda(query.device):
        target_fn("Can't run CUTLASS FMHA; not a CUDA tensor.")
        return False

    device_cc = get_device_cc(query.device)

    if device_cc < 60:
        target_fn(
            "CUTLASS FMHA only supports CUDA devices with compute capability 60 or higher, "
            f"got {device_cc}."
        )
        return False

    head_dim = query.shape[-1]
    head_dim_v = value.shape[-1]

    if query.dtype not in [torch.float32, torch.float16, torch.bfloat16]:
        target_fn(
            "Can't run CUTLASS FMHA; it only supports FP32, FP16, and BF16.",
            exception=ValueError,
        )
        return False

    if head_dim % 8 != 0:
        target_fn(
            "Can't run CUTLASS FMHA; it only supports head dims that are multiples of 8.",
            exception=ValueError,
        )
        return False

    if head_dim_v % 8 != 0:
        target_fn(
            "Can't run CUTLASS FMHA; it only supports head dims that are multiples of 8, "
            f"got {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    if max(head_dim, head_dim_v) > 2**16:
        target_fn(
            f"Can't run CUTLASS FMHA; it supports max head dim of {2**16}, "
            f"got {head_dim=}, {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    return True


def can_run_cutlass_fna(
    query: Tensor, key: Tensor, value: Tensor, raise_error: bool = False
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if not HAS_LIBNATTEN:
        target_fn("Can't run CUTLASS FNA; NATTEN was not built with libnatten.")
        return False

    if not na_tensor_checks(
        query,
        key,
        value,
        must_match_head_dims=False,
        raise_error=raise_error,
        backend_name="CUTLASS FNA",
    ):
        return False

    if query.dim() not in [4, 5, 6]:
        target_fn(
            "CUTLASS FNA expects 4-D, 5-D, or 6-D tensors as inputs (corresponding to NA1D, NA2D, and NA3D), "
            f"got {query.dim()=}.",
            exception=ValueError,
        )
        return False

    if not is_cuda(query.device):
        target_fn("Can't run CUTLASS FNA; not a CUDA tensor.")
        return False

    device_cc = get_device_cc(query.device)

    if device_cc < 60:
        target_fn(
            "CUTLASS FNA only supports CUDA devices with compute capability 60 or higher, "
            f"got {device_cc}."
        )
        return False

    head_dim = query.shape[-1]
    head_dim_v = value.shape[-1]

    if query.dtype not in [torch.float32, torch.float16, torch.bfloat16]:
        target_fn(
            "Can't run CUTLASS FNA; it only supports FP32, FP16, and BF16.",
            exception=ValueError,
        )
        return False

    if head_dim % 8 != 0:
        target_fn(
            "Can't run CUTLASS FNA; it only supports head dims that are multiples of 8, "
            f"got {head_dim=}.",
            exception=ValueError,
        )
        return False

    if head_dim_v % 8 != 0:
        target_fn(
            "Can't run CUTLASS FNA; it only supports head dims that are multiples of 8, "
            f"got {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    if max(head_dim, head_dim_v) > 2**16:
        target_fn(
            f"Can't run CUTLASS FNA; it supports max head dim of {2**16}, "
            f"got {head_dim=}, {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    return True


### Flex FMHA/FNA

_FLEX_SUPPORTED = _TORCH_VERSION >= [2, 7]
_FLEX_COMPILE_SUPPORTED = _TORCH_VERSION >= [2, 7] and _IS_TORCH_COMPILE_SUPPORTED


def can_run_flex_attention(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    torch_compile: bool,
    raise_error: bool = False,
) -> bool:
    target_fn = functools.partial(log_or_raise_error, raise_error=raise_error)

    if not _FLEX_SUPPORTED:
        target_fn("Can't run NATTEN with Flex Attention with torch < 2.7.")
        return False

    if torch_compile and not _FLEX_COMPILE_SUPPORTED:
        target_fn("Can't run NATTEN with Flex Attention (compiled).)")
        return False

    if torch_compile and not is_flex_compile_allowed():
        target_fn(
            "NATTEN does not allow compiling Flex Attention. This is because we cannot verify "
            "Flex's correctness in all scenarios through NATTEN's tests. You can choose to override "
            "this, though it is discouraged, as it may affect your results significantly, "
            "by doing:\n"
            "  from natten import allow_flex_compile\n"
            "  allow_flex_compile()\n"
        )
        return False

    if torch_compile and query.requires_grad and not is_flex_compile_backprop_allowed():
        target_fn(
            "NATTEN does not allow compiling Flex Attention for backpropagation "
            "({q,k,v}.requires_grad=True). This is because we cannot verify Flex's correctness "
            "in all scenarios through NATTEN's tests. You can choose to override this, though "
            "it is HIGHLY discouraged, as it may affect the results of your training significantly, "
            "by doing:\n"
            "  from natten import allow_flex_compile_backprop\n"
            "  allow_flex_compile_backprop()\n"
        )
        return False

    # TODO: can we just have different checks for FMHA vs FNA, like the rest of the backends?
    if query.dim() == 4 and key.dim() == 4 and query.shape[1] != key.shape[1]:
        supported = fmha_tensor_checks(
            query,
            key,
            value,
            must_match_head_dims=True,
            raise_error=raise_error,
            backend_name="Flex FMHA",
        )
    else:
        supported = na_tensor_checks(
            query,
            key,
            value,
            must_match_head_dims=True,
            raise_error=raise_error,
            backend_name="Flex FMHA/FNA",
        )
    if not supported:
        return False

    if query.dim() not in [4, 5, 6]:
        target_fn(
            "Flex backend expects 4-D, 5-D, or 6-D tensors as inputs (corresponding to FMHA/NA1D, "
            f"NA2D, and NA3D), got {query.dim()=}.",
            exception=ValueError,
        )
        return False

    if not is_cuda(query.device):
        if not is_cpu(query.device) and not is_rocm(query.device):
            target_fn(
                "Can't run Flex Attention; tensor is not on a CUDA, ROCm, or CPU device: "
                f"{query.device.type}"
            )

            return False
        # TODO: check if ROCm device supports torch.compile/triton?

    else:
        device_cc = get_device_cc(query.device)

        if device_cc < 70:
            target_fn(
                "Flex Attention (compiled) only supports CUDA devices with compute capability "
                f"70 or higher, got {device_cc}."
            )
            return False

    head_dim = query.shape[-1]
    head_dim_v = value.shape[-1]

    if head_dim != head_dim_v:
        target_fn(
            "Can't run NATTEN with Flex Attention; we don't support different head dims for QK and "
            f"V in this backend yet, got {head_dim=}, {head_dim_v=}.",
            exception=ValueError,
        )
        return False

    if not torch_compile and query.dtype not in [
        torch.float32,
        torch.float16,
        torch.bfloat16,
    ]:
        target_fn(
            "Can't run NATTEN with Flex Attention; we only support FP32, FP16, and BF16 for now.",
            exception=ValueError,
        )
        return False

    if torch_compile and query.dtype not in [torch.float16, torch.bfloat16]:
        target_fn(
            "Can't run NATTEN with Flex Attention (compiled); we only support FP32, FP16, and BF16 for now.",
            exception=ValueError,
        )
        return False

    if torch_compile and (
        head_dim < 32 or head_dim > 512 or not math.log2(head_dim).is_integer()
    ):
        target_fn(
            "Can't run NATTEN with Flex Attention (compiled); we only allow 32 <= head_dim <= 512 "
            f"and only powers of two, got {head_dim}.",
            exception=ValueError,
        )
        return False

    if not torch_compile and (
        head_dim < 8 or head_dim > 512 or not math.log2(head_dim).is_integer()
    ):
        target_fn(
            "Can't run NATTEN with Flex Attention (not compiled); we only allow 8 <= head_dim <= 512 "
            f"and only powers of two, got {head_dim}.",
            exception=ValueError,
        )
        return False

    return True



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################
from typing import Dict, List, Optional, Tuple, Union

import torch
from torch import Tensor

from .attn_merge import merge_attentions_compile, merge_attentions_fn

from .backends import (
    choose_backend,
    choose_fmha_backend,
    cutlass_blackwell_fmha,
    cutlass_blackwell_fna_generic,
    cutlass_fmha,
    cutlass_fna_generic,
    cutlass_hopper_fmha,
    cutlass_hopper_fna_generic,
    flex_fmha,
    flex_fna_generic,
)

from .types import (
    CausalArg1DTypeOrDed,
    CausalArg2DTypeOrDed,
    CausalArg3DTypeOrDed,
    CausalArgTypeOrDed,
    Dimension1DType,
    Dimension1DTypeOrDed,
    Dimension2DType,
    Dimension2DTypeOrDed,
    Dimension3DType,
    Dimension3DTypeOrDed,
    DimensionType,
    DimensionTypeOrDed,
    KernelSchedule,
)
from .utils import log
from .utils.checks import (
    additional_kv_tensor_checks,
    check_all_args,
    check_args_against_input,
    check_kernel_schedule,
    fmha_tensor_checks,
    is_self_attention,
    na_tensor_checks,
)

logger = log.get_logger(__name__)


# Standard Attention


def attention(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    scale: Optional[float] = None,
    backend: Optional[str] = None,
    q_tile_size: Optional[int] = None,
    kv_tile_size: Optional[int] = None,
    backward_q_tile_size: Optional[int] = None,
    backward_kv_tile_size: Optional[int] = None,
    backward_kv_splits: Optional[int] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
    return_lse: bool = False,
) -> Union[Tensor, Tuple[Tensor, Tensor]]:
    """Runs standard dot product attention.

    This operation is used to implement neighborhood cross attention, in which we allow every
    token to interact with some additional context (`additional_keys` and `additional_values`
    tensors in [na1d][natten.na1d], [na2d][natten.na2d], and [na3d][natten.na3d]).
    This operator is also used as a fast path for cases where neighborhood attention is equivalent
    to self attention (not causal along any dims, and `kernel_size` is equal to the number of input
    tokens).

    This operation does not call into PyTorch's SDPA, and only runs one of the NATTEN backends
    (`cutlass-fmha`, `hopper-fmha`, `blackwell-fmha`, `flex-fmha`). Reasons for that include being
    able to control performance-related arguments, return logsumexp, and more.
    For more information refer to [backends](backends.md).

    Parameters:
        query (Tensor): 4-D query tensor, with the heads last layout
            (`[batch, seqlen, heads, head_dim]`)

        key (Tensor): 4-D key tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim]`)

        value (Tensor): 4-D value tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim_v]`)

        scale (float): Attention scale. Defaults to `head_dim ** -0.5`.

    Other Parameters:
        backend (str): Backend implementation to run with. Choices are: `None` (pick the best
            available one), `"cutlass-fmha"`, `"hopper-fmha"`, `"blackwell-fmha"`, `"flex-fmha"`.
            Refer to [backends](backends.md) for more information.

        q_tile_size (int): Tile size along query sequence length in the forward pass kernel.
            You can use [profiler](profiler.md) to find valid choices for your use case.

        kv_tile_size (int): Tile size along key/value sequence length in the forward pass kernel.
            You can use [profiler](profiler.md) to find valid choices for your use case.

        backward_q_tile_size (int): Tile size along query sequence length in the backward pass
            kernel. This is only respected by the `"cutlass-fmha"`, `"hopper-fmha"`, and
            `"blackwell-fmha"` backends.
            You can use [profiler](profiler.md) to find valid choices for your use case.

        backward_kv_tile_size (int): Tile size along key/value sequence length in the backward pass
            kernel. This is only respected by the `"cutlass-fmha"`, `"hopper-fmha"`, and
            `"blackwell-fmha"` backends.
            You can use [profiler](profiler.md) to find valid choices for your use case.

        backward_kv_splits (int): Number of key/value tiles allowed to work in parallel in the
            backward pass kernel. This is only respected by the `"cutlass-fmha"` backend, only when
            [KV parallelism](context.md#kv-parallelism-in-fna) is enabled.

        backward_use_pt_reduction (bool): Whether to use PyTorch eager for computing the `dO * O`
            product required by the backward pass, over the CUTLASS kernel. This only applies to
            the `"cutlass-fmha"` backend.

        run_persistent_kernel (bool): Whether to use persistent tile scheduling in the forward pass
            kernel. This only applies to the `"blackwell-fmha"` backend.

        kernel_schedule (Optional[str]): Kernel type (Hopper architecture only). Choices are
            `None`: pick the default, `"non"` (non-persistent), `"coop"` (warp-specialized
            cooperative), or `"pp"` (warp-specialized ping-ponging). Refer to
            [Hopper FMHA/FNA backend](backends.md#hopper-fna-fmha) for more information.

        torch_compile (bool): Applies only to the `"flex-fmha"` backend. Whether or not to JIT
            compile the attention kernel. Due to this being an experimental feature in PyTorch, we
            do not recommend it, and it is guarded by context flags. Read more in
            [Flex Attention + `torch.compile`](context.md#flex-attention-torchcompile).

        return_lse (bool): Whether or not to return the `logsumexp` tensor. `logsumexp` can be used
            in the backward pass, and for [attention merging][natten.merge_attentions].

    Returns:
        output (Tensor): 4-D output tensor, with the heads last layout
            (`[batch, seqlen, heads, head_dim_v]`).

        logsumexp (Tensor): only returned when `return_lse=True`. 3-D logsumexp tensor, with the
            heads last layout (`[batch, seqlen, heads]`).
    """

    fmha_tensor_checks(query, key, value)

    scale = scale or query.shape[-1] ** -0.5

    kernel_schedule = check_kernel_schedule(kernel_schedule)

    backend = backend or choose_fmha_backend(
        query, key, value, torch_compile=torch_compile
    )

    if backend == "blackwell-fmha":
        return cutlass_blackwell_fmha(
            query=query,
            key=key,
            value=value,
            scale=scale,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            backward_q_tile_size=backward_q_tile_size,
            backward_kv_tile_size=backward_kv_tile_size,
            run_persistent_kernel=run_persistent_kernel,
            return_lse=return_lse,
        )

    if backend == "hopper-fmha":
        return cutlass_hopper_fmha(
            query=query,
            key=key,
            value=value,
            scale=scale,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            backward_q_tile_size=backward_q_tile_size,
            backward_kv_tile_size=backward_kv_tile_size,
            kernel_schedule=kernel_schedule,
            return_lse=return_lse,
        )

    elif backend == "cutlass-fmha":
        return cutlass_fmha(
            query=query,
            key=key,
            value=value,
            scale=scale,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            backward_q_tile_size=backward_q_tile_size,
            backward_kv_tile_size=backward_kv_tile_size,
            backward_kv_splits=backward_kv_splits,
            backward_use_pt_reduction=backward_use_pt_reduction,
            return_lse=return_lse,
        )

    elif backend == "flex-fmha":
        return flex_fmha(
            query=query,
            key=key,
            value=value,
            scale=scale,
            q_tile_size=q_tile_size,
            kv_tile_size=kv_tile_size,
            torch_compile=torch_compile,
            return_lse=return_lse,
        )

    raise NotImplementedError(f"Unrecognized NATTEN FMHA backend {backend}.")


def merge_attentions(
    outputs: List[Tensor], lse_tensors: List[Tensor], torch_compile: bool = True
) -> Tensor:
    """Takes multiple attention *outputs* originating from the same query tensor, and their
    corresponding logsumexps, and merges them as if their context (key/value pair) had been
    concatenated.

    This operation is used to implement cross-neighborhood attention, and can also be used for
    distributed setups, such as context-parallelism.

    This operation also attempts to use `torch.compile` to fuse the elementwise operations. This
    can be disabled by passing `torch_compile=False`.

    Args:
        outputs (List[Tensor]): List of 4-D attention output tensors, with the heads last layout
            (`[batch, seqlen, heads, head_dim]`)

        lse_tensors (List[Tensor]): List of 3-D logsumexp tensors, with the heads last layout
            (`[batch, seqlen, heads]`)

        torch_compile (bool): Attempt to use `torch.compile` to fuse the underlying elementwise
            operations.

    Returns:
        output (Tensor): merged attention output.
    """

    if len(outputs) < 2:
        raise ValueError("`merge_attentions` expects at least two tensors.")

    if len(outputs) != len(lse_tensors):
        raise ValueError(
            "`merge_attentions` expected number of outputs and LSE tensors to match, "
            f"got {len(outputs)=} != {len(lse_tensors)}."
        )

    for i, (output, lse) in enumerate(zip(outputs, lse_tensors)):
        if output.dim() != 4 or not output.is_contiguous():
            raise ValueError(
                "Output tensors must be rank-4 tensors with (batch, seq, heads, dim), "
                f"but got output {i} with rank={output.dim()}."
            )

        if output.shape != outputs[0].shape:
            raise ValueError(
                f"Output tensors must must match in shape, but got output {i} "
                f"with shape={output.shape}."
            )

        if lse.dim() != 3:
            raise ValueError(
                "LSE tensors must be rank-3 tensors with (batch, seq, heads)"
                f"but got LSE {i} with rank={lse.dim()}."
            )

        if lse.shape != outputs[0].shape[:3]:
            raise ValueError(
                f"LSE tensors must must match outputs in shape except last dim "
                f"({outputs[0].shape=}), but got LSE {i} with shape={lse.shape}."
            )

    if not torch_compile:
        return merge_attentions_fn(
            [output.contiguous() for output in outputs],
            [lse.contiguous() for lse in lse_tensors],
        )

    return merge_attentions_compile(
        [output.contiguous() for output in outputs],
        [lse.contiguous() for lse in lse_tensors],
    )


# Neighborhood Attention


def neighborhood_attention_generic(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: DimensionTypeOrDed,
    stride: DimensionTypeOrDed = 1,
    dilation: DimensionTypeOrDed = 1,
    is_causal: Optional[CausalArgTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    attention_kwargs: Optional[Dict] = None,
    # Perf-related args
    backend: Optional[str] = None,
    q_tile_shape: Optional[DimensionType] = None,
    kv_tile_shape: Optional[DimensionType] = None,
    backward_q_tile_shape: Optional[DimensionType] = None,
    backward_kv_tile_shape: Optional[DimensionType] = None,
    backward_kv_splits: Optional[DimensionType] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
) -> Tensor:

    na_tensor_checks(query, key, value)
    additional_kv_tensor_checks(query, key, value, additional_keys, additional_values)
    kernel_schedule = check_kernel_schedule(kernel_schedule)

    na_dim = query.dim() - 3  # batch, heads, head_dim

    assert na_dim in [1, 2, 3]

    kernel_size, stride, dilation, is_causal = check_all_args(
        na_dim, kernel_size, stride, dilation, is_causal
    )

    check_args_against_input(
        query,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
    )

    if is_self_attention(query, kernel_size=kernel_size, is_causal=is_causal):
        logger.debug(
            f"{query.shape=} with {kernel_size=} and {is_causal=} is self attention. "
            "Calling attention instead of neighborhood attention directly."
        )

        query_shape = query.shape
        query = query.flatten(1, na_dim)
        key = key.flatten(1, na_dim)
        value = value.flatten(1, na_dim)

        if additional_keys is not None and additional_values is not None:
            key = torch.cat([key, additional_keys], dim=1)
            value = torch.cat([value, additional_values], dim=1)

        attn_kwargs = attention_kwargs or {}
        out: Tensor = attention(  # type: ignore[assignment]
            query,
            key,
            value,
            scale=scale,
            return_lse=False,
            **attn_kwargs,
        )
        output_shape = [s for s in query_shape[:-1]] + [value.shape[-1]]
        return out.reshape(*output_shape)

    scale = scale or query.shape[-1] ** -0.5

    backend = backend or choose_backend(query, key, value, torch_compile=torch_compile)

    has_additional_attention = (
        additional_keys is not None and additional_values is not None
    )

    if backend == "blackwell-fna":
        outputs = cutlass_blackwell_fna_generic(
            query=query,
            key=key,
            value=value,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            scale=scale,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            backward_q_tile_shape=backward_q_tile_shape,
            backward_kv_tile_shape=backward_kv_tile_shape,
            run_persistent_kernel=run_persistent_kernel,
            return_lse=has_additional_attention,
        )

    elif backend == "hopper-fna":
        outputs = cutlass_hopper_fna_generic(
            query=query,
            key=key,
            value=value,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            scale=scale,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            backward_q_tile_shape=backward_q_tile_shape,
            backward_kv_tile_shape=backward_kv_tile_shape,
            kernel_schedule=kernel_schedule,
            return_lse=has_additional_attention,
        )

    elif backend == "cutlass-fna":
        outputs = cutlass_fna_generic(
            query=query,
            key=key,
            value=value,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            scale=scale,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            backward_q_tile_shape=backward_q_tile_shape,
            backward_kv_tile_shape=backward_kv_tile_shape,
            backward_kv_splits=backward_kv_splits,
            backward_use_pt_reduction=backward_use_pt_reduction,
            return_lse=has_additional_attention,
        )

    elif backend == "flex-fna":
        outputs = flex_fna_generic(
            query=query,
            key=key,
            value=value,
            kernel_size=kernel_size,
            stride=stride,
            dilation=dilation,
            is_causal=is_causal,
            scale=scale,
            q_tile_shape=q_tile_shape,
            kv_tile_shape=kv_tile_shape,
            torch_compile=torch_compile,
            return_lse=has_additional_attention,
        )

    else:
        raise NotImplementedError(f"Unrecognized NATTEN backend {backend}.")

    if has_additional_attention:
        assert additional_keys is not None
        assert additional_values is not None
        assert outputs is not None and isinstance(outputs, tuple) and len(outputs) == 2
        out, lse = outputs

        attention_kwargs = attention_kwargs or {}
        additional_output, additional_lse = attention(
            query.flatten(1, na_dim),
            additional_keys,
            additional_values,
            scale=scale,
            return_lse=True,
            **attention_kwargs,
        )

        merged_output = merge_attentions(
            [out.flatten(1, na_dim), additional_output],
            [lse.flatten(1, na_dim), additional_lse],
        )
        return merged_output.reshape(out.shape)

    assert isinstance(outputs, Tensor)

    return outputs


def na1d(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension1DTypeOrDed,
    stride: Dimension1DTypeOrDed = 1,
    dilation: Dimension1DTypeOrDed = 1,
    is_causal: Optional[CausalArg1DTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    attention_kwargs: Optional[Dict] = None,
    backend: Optional[str] = None,
    q_tile_shape: Optional[Dimension1DType] = None,
    kv_tile_shape: Optional[Dimension1DType] = None,
    backward_q_tile_shape: Optional[Dimension1DType] = None,
    backward_kv_tile_shape: Optional[Dimension1DType] = None,
    backward_kv_splits: Optional[Dimension1DType] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
) -> Tensor:
    """Computes 1-D neighborhood attention.

    Args:
        query (Tensor): 4-D query tensor, with the heads last layout
            (`[batch, seqlen, heads, head_dim]`)

        key (Tensor): 4-D key tensor, with the heads last layout
            (`[batch, seqlen, heads, head_dim]`)

        value (Tensor): 4-D value tensor, with the heads last layout
            (`[batch, seqlen, heads, head_dim]`)

        kernel_size (Tuple[int] | int): Neighborhood window (kernel) size.

            !!! note
                `kernel_size` must be smaller than or equal to `seqlen`.

        stride (Tuple[int] | int): Sliding window step size. Defaults to `1` (standard sliding
            window).

            !!! note
                `stride` must be smaller than or equal to `kernel_size`.
                When `stride == kernel_size`, there will be no overlap between sliding windows,
                which is equivalent to blocked attention (a.k.a.
                [window self attention](https://arxiv.org/abs/2103.14030)).

        dilation (Tuple[int] | int): Dilation step size. Defaults to `1` (standard sliding window).

            !!! note
                The product of `dilation` and `kernel_size` must be smaller than or equal to
                `seqlen`.

        is_causal (Tuple[bool] | bool): Toggle causal masking. Defaults to `False`
            (bi-directional).

        scale (float): Attention scale. Defaults to `head_dim ** -0.5`.

        additional_keys: `None` or 4-D key tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim]`), corresponding to key tokens from some
            additional context. Used when performing neighborhood cross-attention, where `query`
            tokens attend to their neighborhood, as well as some fixed additional set of tokens.

        additional_values: `None` or 4-D value tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim]`), corresponding to value tokens from some
            additional context. Used when performing neighborhood cross-attention, where `query`
            tokens attend to their neighborhood, as well as some fixed additional set of tokens.

            !!! note
                `additional_keys` and `additional_values` must both either be `Tensor`s, or both
                `None`s, and must match in shape.

    Other Parameters:
        backend (str): Backend implementation to run with. Choices are: `None` (pick the best
            available one), `"cutlass-fna"`, `"hopper-fna"`, `"blackwell-fna"`, `"flex-fna"`.
            Refer to [backends](backends.md) for more information.

        q_tile_shape (Tuple[int]): 1-D Tile shape for the query token layout in the forward pass
            kernel. You can use [profiler](profiler.md) to find valid choices for your use case,
            and search for the best combination.

        kv_tile_shape (Tuple[int]): 1-D Tile shape for the key-value token layout in the forward
            pass kernel. You can use [profiler](profiler.md) to find valid choices for your use
            case, and search for the best combination.

        backward_q_tile_shape (Tuple[int]): 1-D Tile shape for the query token layout in the
            backward pass kernel. This is only respected by the `"cutlass-fna"` backend.
            You can use [profiler](profiler.md) to find valid choices for your use case, and
            search for the best combination.

        backward_kv_tile_shape (Tuple[int]): 1-D Tile shape for the key/value token layout in the
            backward pass kernel. This is only respected by the `"cutlass-fna"` backend.
            You can use [profiler](profiler.md) to find valid choices for your use case, and
            search for the best combination.

        backward_kv_splits (Tuple[int]): Number of key/value tiles allowed to work in parallel in
            the backward pass kernel. Like tile shapes, this is a tuple and not an integer for
            neighborhood attention operations, and the size of the tuple corresponds to the number
            of dimensions / rank of the layout of tokens. This is only respected by the
            `"cutlass-fna"` backend, and only when
            [KV parallelism](context.md#kv-parallelism-in-fna) is enabled.

        backward_use_pt_reduction (bool): Whether to use PyTorch eager for computing the `dO * O`
            product required by the backward pass, over the CUTLASS kernel. This only applies to
            the `"cutlass-fna"` backend.

        run_persistent_kernel (bool): Whether to use persistent tile scheduling in the forward pass
            kernel. This only applies to the `"blackwell-fna"` backend.

        kernel_schedule (Optional[str]): Kernel type (Hopper architecture only). Choices are
            `None`: pick the default, `"non"` (non-persistent), `"coop"` (warp-specialized
            cooperative), or `"pp"` (warp-specialized ping-ponging). Refer to
            [Hopper FMHA/FNA backend](backends.md#hopper-fna-fmha) for more information.

        torch_compile (bool): Applies only to the `"flex-fna"` backend. Whether or not to JIT
            compile the attention kernel. Due to this being an experimental feature in PyTorch, we
            do not recommend it, and it is guarded by context flags. Read more in
            [Flex Attention + `torch.compile`](context.md#flex-attention-torchcompile).

        attention_kwargs: arguments to the [attention][natten.attention] operator, if used to
            implement neighborhood cross-attention, or self attention as a fast path for
            neighborhood attention.

            If `additional_{keys,values}` are specified, NATTEN usually performs a separate
            cross-attention using our [attention][natten.attention] operator, and
            [merges][natten.merge_attentions] the results.

            If for a given use case, the neighborhood attention problem is equivalent to self
            attention (not causal, `kernel_size == seqlen`), NATTEN will also attempt to directly
            use [attention][natten.attention].

            You can override arguments to [attention][natten.attention] by passing a
            dictionary here.

            !!! example
                ```python
                out = na1d(
                    q, k, v, kernel_size=kernel_size,
                    ...,
                    attention_kwargs={
                        "backend": "blackwell-fmha",
                        "run_persistent_kernel": True,
                    }
                )
                ```

    Returns:
        output (Tensor): 4-D output tensor, with the heads last layout
            (`[batch, seqlen, heads, head_dim]`).
    """
    return neighborhood_attention_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        additional_keys=additional_keys,
        additional_values=additional_values,
        attention_kwargs=attention_kwargs,
        backend=backend,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        backward_kv_splits=backward_kv_splits,
        backward_use_pt_reduction=backward_use_pt_reduction,
        run_persistent_kernel=run_persistent_kernel,
        kernel_schedule=kernel_schedule,
        torch_compile=torch_compile,
    )


def na2d(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension2DTypeOrDed,
    stride: Dimension2DTypeOrDed = 1,
    dilation: Dimension2DTypeOrDed = 1,
    is_causal: Optional[CausalArg2DTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    attention_kwargs: Optional[Dict] = None,
    backend: Optional[str] = None,
    q_tile_shape: Optional[Dimension2DType] = None,
    kv_tile_shape: Optional[Dimension2DType] = None,
    backward_q_tile_shape: Optional[Dimension2DType] = None,
    backward_kv_tile_shape: Optional[Dimension2DType] = None,
    backward_kv_splits: Optional[Dimension2DType] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
) -> Tensor:
    """Computes 2-D neighborhood attention.

    Args:
        query (Tensor): 2-D query tensor, with the heads last layout:
            `[batch, X, Y, heads, head_dim]`, where token layout shape (feature map shape) is
            `(X, Y)`.

        key (Tensor): 2-D key tensor, with the heads last layout:
            `[batch, X, Y, heads, head_dim]`, where token layout shape (feature map shape) is
            `(X, Y)`.

        value (Tensor): 2-D value tensor, with the heads last layout:
            `[batch, X, Y, heads, head_dim]`, where token layout shape (feature map shape) is
            `(X, Y)`.

        kernel_size (Tuple[int, int] | int): Neighborhood window (kernel) size/shape. If an
            integer, it will be repeated for all 2 dimensions. For example `kernel_size=3` is
            reinterpreted as `kernel_size=(3, 3)`.

            !!! note
                `kernel_size` must be smaller than or equal to token layout shape (`(X, Y)`) along
                every dimension.

        stride (Tuple[int, int] | int): Sliding window step size/shape. Defaults to `1` (standard
            sliding window). If an integer, it will be repeated for all 2 dimensions. For example
            `stride=2` is reinterpreted as `stride=(2, 2)`.

            !!! note
                `stride` must be smaller than or equal to `kernel_size` along every dimension.
                When `stride == kernel_size`, there will be no overlap between sliding windows,
                which is equivalent to blocked attention (a.k.a.
                [window self attention](https://arxiv.org/abs/2103.14030)).

        dilation (Tuple[int, int] | int): Dilation step size/shape. Defaults to `1` (standard
            sliding window). If an integer, it will be repeated for all 2 dimensions. For example
            `dilation=4` is reinterpreted as `dilation=(4, 4)`.

            !!! note
                The product of `dilation` and `kernel_size` must be smaller than or equal to
                token layout shape (`(X, Y)`) along every dimension.

        is_causal (Tuple[bool, bool] | bool): Toggle causal masking. Defaults to `False`
            (bi-directional). If a boolean, it will be repeated for all 2 dimensions. For example
            `is_causal=True` is reinterpreted as `is_causal=(True, True)`.

        scale (float): Attention scale. Defaults to `head_dim ** -0.5`.

        additional_keys: `None` or 4-D key tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim]`), corresponding to key tokens from some
            additional context. Used when performing neighborhood cross-attention, where `query`
            tokens attend to their neighborhood, as well as some fixed additional set of tokens.

        additional_values: `None` or 4-D value tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim]`), corresponding to value tokens from some
            additional context. Used when performing neighborhood cross-attention, where `query`
            tokens attend to their neighborhood, as well as some fixed additional set of tokens.

            !!! note
                `additional_keys` and `additional_values` must both either be `Tensor`s, or both
                `None`s, and must match in shape.

    Other Parameters:
        backend (str): Backend implementation to run with. Choices are: `None` (pick the best
            available one), `"cutlass-fna"`, `"hopper-fna"`, `"blackwell-fna"`, `"flex-fna"`.
            Refer to [backends](backends.md) for more information.

        q_tile_shape (Tuple[int, int]): 2-D Tile shape for the query token layout in the forward
            pass kernel. You can use [profiler](profiler.md) to find valid choices for your use
            case, and search for the best combination.

        kv_tile_shape (Tuple[int, int]): 2-D Tile shape for the key-value token layout in the
            forward pass kernel. You can use [profiler](profiler.md) to find valid choices for your
            use case, and search for the best combination.

        backward_q_tile_shape (Tuple[int, int]): 2-D Tile shape for the query token layout in the
            backward pass kernel. This is only respected by the `"cutlass-fna"` backend.
            You can use [profiler](profiler.md) to find valid choices for your use case, and
            search for the best combination.

        backward_kv_tile_shape (Tuple[int, int]): 2-D Tile shape for the key/value token layout in
            the backward pass kernel. This is only respected by the `"cutlass-fna"` backend.
            You can use [profiler](profiler.md) to find valid choices for your use case, and
            search for the best combination.

        backward_kv_splits (Tuple[int, int]): Number of key/value tiles allowed to work in parallel
            in the backward pass kernel. Like tile shapes, this is a tuple and not an integer for
            neighborhood attention operations, and the size of the tuple corresponds to the number
            of dimensions / rank of the layout of tokens. This is only respected by the
            `"cutlass-fna"` backend, and only when
            [KV parallelism](context.md#kv-parallelism-in-fna) is enabled.

        backward_use_pt_reduction (bool): Whether to use PyTorch eager for computing the `dO * O`
            product required by the backward pass, over the CUTLASS kernel. This only applies to
            the `"cutlass-fna"` backend.

        run_persistent_kernel (bool): Whether to use persistent tile scheduling in the forward pass
            kernel. This only applies to the `"blackwell-fna"` backend.

        kernel_schedule (Optional[str]): Kernel type (Hopper architecture only). Choices are
            `None`: pick the default, `"non"` (non-persistent), `"coop"` (warp-specialized
            cooperative), or `"pp"` (warp-specialized ping-ponging). Refer to
            [Hopper FMHA/FNA backend](backends.md#hopper-fna-fmha) for more information.

        torch_compile (bool): Applies only to the `"flex-fna"` backend. Whether or not to JIT
            compile the attention kernel. Due to this being an experimental feature in PyTorch, we
            do not recommend it, and it is guarded by context flags. Read more in
            [Flex Attention + `torch.compile`](context.md#flex-attention-torchcompile).

        attention_kwargs: arguments to the [attention][natten.attention] operator, if used to
            implement neighborhood cross-attention, or self attention as a fast path for
            neighborhood attention.

            If `additional_{keys,values}` are specified, NATTEN usually performs a separate
            cross-attention using our [attention][natten.attention] operator, and
            [merges][natten.merge_attentions] the results.

            If for a given use case, the neighborhood attention problem is equivalent to self
            attention (not causal along any dims, `kernel_size == (X, Y)`), NATTEN will also
            attempt to directly use [attention][natten.attention].

            You can override arguments to [attention][natten.attention] by passing a
            dictionary here.

            !!! example
                ```python
                out = na2d(
                    q, k, v, kernel_size=kernel_size,
                    ...,
                    attention_kwargs={
                        "backend": "blackwell-fmha",
                        "run_persistent_kernel": True,
                    }
                )
                ```

    Returns:
        output (Tensor): 5-D output tensor, with the heads last layout
            (`[batch, X, Y, heads, head_dim]`).
    """
    return neighborhood_attention_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        additional_keys=additional_keys,
        additional_values=additional_values,
        attention_kwargs=attention_kwargs,
        backend=backend,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        backward_kv_splits=backward_kv_splits,
        backward_use_pt_reduction=backward_use_pt_reduction,
        run_persistent_kernel=run_persistent_kernel,
        kernel_schedule=kernel_schedule,
        torch_compile=torch_compile,
    )


def na3d(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    kernel_size: Dimension3DTypeOrDed,
    stride: Dimension3DTypeOrDed = 1,
    dilation: Dimension3DTypeOrDed = 1,
    is_causal: Optional[CausalArg3DTypeOrDed] = False,
    scale: Optional[float] = None,
    additional_keys: Optional[Tensor] = None,
    additional_values: Optional[Tensor] = None,
    attention_kwargs: Optional[Dict] = None,
    backend: Optional[str] = None,
    q_tile_shape: Optional[Dimension3DType] = None,
    kv_tile_shape: Optional[Dimension3DType] = None,
    backward_q_tile_shape: Optional[Dimension3DType] = None,
    backward_kv_tile_shape: Optional[Dimension3DType] = None,
    backward_kv_splits: Optional[Dimension3DType] = None,
    backward_use_pt_reduction: bool = False,
    run_persistent_kernel: bool = True,
    kernel_schedule: Optional[Union[str, KernelSchedule]] = None,
    torch_compile: bool = False,
) -> Tensor:
    """Computes 3-D neighborhood attention.

    Args:
        query (Tensor): 3-D query tensor, with the heads last layout:
            `[batch, X, Y, Z, heads, head_dim]`, where token layout shape (feature map shape) is
            `(X, Y, Z)`.

        key (Tensor): 3-D key tensor, with the heads last layout:
            `[batch, X, Y, Z, heads, head_dim]`, where token layout shape (feature map shape) is
            `(X, Y, Z)`.

        value (Tensor): 3-D value tensor, with the heads last layout:
            `[batch, X, Y, Z, heads, head_dim]`, where token layout shape (feature map shape) is
            `(X, Y, Z)`.

        kernel_size (Tuple[int, int, int] | int): Neighborhood window (kernel) size/shape. If an
            integer, it will be repeated for all 3 dimensions. For example `kernel_size=3` is
            reinterpreted as `kernel_size=(3, 3, 3)`.

            !!! note
                `kernel_size` must be smaller than or equal to token layout shape (`(X, Y, Z)`)
                along every dimension.

        stride (Tuple[int, int, int] | int): Sliding window step size/shape. Defaults to `1`
            (standard sliding window). If an integer, it will be repeated for all 3 dimensions.
            For example `stride=2` is reinterpreted as `stride=(2, 2, 2)`.

            !!! note
                `stride` must be smaller than or equal to `kernel_size` along every dimension.
                When `stride == kernel_size`, there will be no overlap between sliding windows,
                which is equivalent to blocked attention (a.k.a.
                [window self attention](https://arxiv.org/abs/2103.14030)).

        dilation (Tuple[int, int, int] | int): Dilation step size/shape. Defaults to `1` (standard
            sliding window). If an integer, it will be repeated for all 3 dimensions. For example
            `dilation=4` is reinterpreted as `dilation=(4, 4, 4)`.

            !!! note
                The product of `dilation` and `kernel_size` must be smaller than or equal to
                token layout shape (`(X, Y, Z)`) along every dimension.

        is_causal (Tuple[bool, bool, bool] | bool): Toggle causal masking. Defaults to `False`
            (bi-directional). If a boolean, it will be repeated for all 3 dimensions. For example
            `is_causal=True` is reinterpreted as `is_causal=(True, True, True)`.

        scale (float): Attention scale. Defaults to `head_dim ** -0.5`.

        additional_keys: `None` or 4-D key tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim]`), corresponding to key tokens from some
            additional context. Used when performing neighborhood cross-attention, where `query`
            tokens attend to their neighborhood, as well as some fixed additional set of tokens.

        additional_values: `None` or 4-D value tensor, with the heads last layout
            (`[batch, seqlen_kv, heads, head_dim]`), corresponding to value tokens from some
            additional context. Used when performing neighborhood cross-attention, where `query`
            tokens attend to their neighborhood, as well as some fixed additional set of tokens.

            !!! note
                `additional_keys` and `additional_values` must both either be `Tensor`s, or both
                `None`s, and must match in shape.

    Other Parameters:
        backend (str): Backend implementation to run with. Choices are: `None` (pick the best
            available one), `"cutlass-fna"`, `"hopper-fna"`, `"blackwell-fna"`, `"flex-fna"`.
            Refer to [backends](backends.md) for more information.

        q_tile_shape (Tuple[int, int, int]): 3-D Tile shape for the query token layout in the
            forward pass kernel. You can use [profiler](profiler.md) to find valid choices for your
            use case, and search for the best combination.

        kv_tile_shape (Tuple[int, int, int]): 3-D Tile shape for the key-value token layout in the
            forward pass kernel. You can use [profiler](profiler.md) to find valid choices for your
            use case, and search for the best combination.

        backward_q_tile_shape (Tuple[int, int, int]): 3-D Tile shape for the query token layout in
            the backward pass kernel. This is only respected by the `"cutlass-fna"` backend.
            You can use [profiler](profiler.md) to find valid choices for your use case, and
            search for the best combination.

        backward_kv_tile_shape (Tuple[int, int, int]): 3-D Tile shape for the key/value token
            layout in the backward pass kernel. This is only respected by the `"cutlass-fna"`
            backend. You can use [profiler](profiler.md) to find valid choices for your use case,
            and search for the best combination.

        backward_kv_splits (Tuple[int, int, int]): Number of key/value tiles allowed to work in
            parallel in the backward pass kernel. Like tile shapes, this is a tuple and not an
            integer for neighborhood attention operations, and the size of the tuple corresponds to
            the number of dimensions / rank of the layout of tokens. This is only respected by the
            `"cutlass-fna"` backend, and only when
            [KV parallelism](context.md#kv-parallelism-in-fna) is enabled.

        backward_use_pt_reduction (bool): Whether to use PyTorch eager for computing the `dO * O`
            product required by the backward pass, over the CUTLASS kernel. This only applies to
            the `"cutlass-fna"` backend.

        run_persistent_kernel (bool): Whether to use persistent tile scheduling in the forward pass
            kernel. This only applies to the `"blackwell-fna"` backend.

        kernel_schedule (Optional[str]): Kernel type (Hopper architecture only). Choices are
            `None`: pick the default, `"non"` (non-persistent), `"coop"` (warp-specialized
            cooperative), or `"pp"` (warp-specialized ping-ponging). Refer to
            [Hopper FMHA/FNA backend](backends.md#hopper-fna-fmha) for more information.

        torch_compile (bool): Applies only to the `"flex-fna"` backend. Whether or not to JIT
            compile the attention kernel. Due to this being an experimental feature in PyTorch, we
            do not recommend it, and it is guarded by context flags. Read more in
            [Flex Attention + `torch.compile`](context.md#flex-attention-torchcompile).

        attention_kwargs: arguments to the [attention][natten.attention] operator, if used to
            implement neighborhood cross-attention, or self attention as a fast path for
            neighborhood attention.

            If `additional_{keys,values}` are specified, NATTEN usually performs a separate
            cross-attention using our [attention][natten.attention] operator, and
            [merges][natten.merge_attentions] the results.

            If for a given use case, the neighborhood attention problem is equivalent to self
            attention (not causal along any dims, `kernel_size == (X, Y, Z)`), NATTEN will also
            attempt to directly use [attention][natten.attention].

            You can override arguments to [attention][natten.attention] by passing a
            dictionary here.

            !!! example
                ```python
                out = na3d(
                    q, k, v, kernel_size=kernel_size,
                    ...,
                    attention_kwargs={
                        "backend": "blackwell-fmha",
                        "run_persistent_kernel": True,
                    }
                )
                ```

    Returns:
        output (Tensor): 6-D output tensor, with the heads last layout
            (`[batch, X, Y, Z, heads, head_dim]`).
    """
    return neighborhood_attention_generic(
        query=query,
        key=key,
        value=value,
        kernel_size=kernel_size,
        stride=stride,
        dilation=dilation,
        is_causal=is_causal,
        scale=scale,
        additional_keys=additional_keys,
        additional_values=additional_values,
        attention_kwargs=attention_kwargs,
        backend=backend,
        q_tile_shape=q_tile_shape,
        kv_tile_shape=kv_tile_shape,
        backward_q_tile_shape=backward_q_tile_shape,
        backward_kv_tile_shape=backward_kv_tile_shape,
        backward_kv_splits=backward_kv_splits,
        backward_use_pt_reduction=backward_use_pt_reduction,
        run_persistent_kernel=run_persistent_kernel,
        kernel_schedule=kernel_schedule,
        torch_compile=torch_compile,
    )



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################


from typing import List

from ...utils import log

logger = log.get_logger(__name__)

import torch  # noqa: F401
from torch import Tensor

from ...types import (
    CutlassBlackwellFmhaBackwardConfigType,
    CutlassBlackwellFmhaForwardConfigType,
    CutlassBlackwellFnaBackwardConfigType,
    CutlassBlackwellFnaForwardConfigType,
    CutlassFmhaForwardConfigType,
    CutlassFnaForwardConfigType,
    CutlassHopperFmhaBackwardConfigType,
    CutlassHopperFmhaForwardConfigType,
    CutlassHopperFnaBackwardConfigType,
    CutlassHopperFnaForwardConfigType,
    FlexFmhaForwardConfigType,
    FlexFnaForwardConfigType,
)

from .checks import (
    can_run_cutlass_blackwell_fmha,
    can_run_cutlass_blackwell_fna,
    can_run_cutlass_fmha,
    can_run_cutlass_fna,
    can_run_cutlass_hopper_fmha,
    can_run_cutlass_hopper_fna,
    can_run_flex_attention,
)

from .cutlass import (
    get_all_tile_shapes_backward as get_all_cutlass_fna_backward_configs,
    get_all_tile_shapes_forward as get_all_cutlass_fna_forward_configs,
    get_all_tile_sizes_backward as get_all_cutlass_fmha_backward_configs,
    get_all_tile_sizes_forward as get_all_cutlass_fmha_forward_configs,
)

from .cutlass_blackwell import (
    get_all_backward_configs as get_all_blackwell_fna_backward_configs,
    get_all_fmha_backward_configs as get_all_blackwell_fmha_backward_configs,
    get_all_fmha_forward_configs as get_all_blackwell_fmha_forward_configs,
    get_all_forward_configs as get_all_blackwell_fna_forward_configs,
)

from .cutlass_hopper import (
    get_all_backward_configs as get_all_hopper_fna_backward_configs,
    get_all_fmha_backward_configs as get_all_hopper_fmha_backward_configs,
    get_all_fmha_forward_configs as get_all_hopper_fmha_forward_configs,
    get_all_forward_configs as get_all_hopper_fna_forward_configs,
)

from .flex import (
    get_all_tile_shapes_forward as get_all_flex_fna_forward_configs,
    get_all_tile_sizes_forward as get_all_flex_fmha_forward_configs,
)


### CUTLASS Blackwell kernels


def get_configs_for_cutlass_blackwell_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassBlackwellFmhaForwardConfigType]:
    """Returns Blackwell FMHA configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Blackwell datacenter GPU (SM100; compute capability
    10.0), and if so, returns *forward pass* configurations compatible with the tensor dtype and
    head dim.

    Each configuration for this operation is a tuple of two integers: `(q_tile_size,
    kv_tile_size)`. These are arguments to [natten.attention][natten.attention].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[int, int]]): List of tuples of two integers corresponding to query and KV tile
            sizes.
    """

    if not can_run_cutlass_blackwell_fmha(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_blackwell_fmha_forward_configs(input_tensor=query)


def get_bwd_configs_for_cutlass_blackwell_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassBlackwellFmhaBackwardConfigType]:
    """Returns Blackwell FMHA backward pass configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Blackwell datacenter GPU (SM100; compute capability
    10.0), and if so, returns *backward pass* configurations compatible with the tensor dtype and
    head dim.

    Each configuration for this operation is a tuple of two integers: `(q_tile_size,
    kv_tile_size)`. These are arguments to [natten.attention][natten.attention].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[int, int]]): List of tuples of two integers corresponding to query and KV tile
            sizes.
    """

    if not can_run_cutlass_blackwell_fmha(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_blackwell_fmha_backward_configs(input_tensor=query)


def get_configs_for_cutlass_blackwell_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassBlackwellFnaForwardConfigType]:
    """Returns Blackwell FNA configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Blackwell datacenter GPU (SM100; compute capability
    10.0), and if so, returns *forward pass* configurations compatible with the tensor dtype and
    head dim, and according to the rank of the token layout (1D/2D/3D).

    Each configuration for this operation is a tuple of two integer tuples: `(q_tile_shape,
    kv_tile_shape)`. These are arguments to [natten.na1d][natten.na1d], [natten.na2d][natten.na2d],
    and [natten.na3d][natten.na3d].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[tuple, tuple]]): List of tuples of two integer tuples corresponding to query
            and KV tile *shapes*.
    """
    if not can_run_cutlass_blackwell_fna(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_blackwell_fna_forward_configs(input_tensor=query)


def get_bwd_configs_for_cutlass_blackwell_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassBlackwellFnaBackwardConfigType]:
    """Returns Blackwell FNA backward pass configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Blackwell datacenter GPU (SM100; compute capability
    10.0), and if so, returns *backward pass* configurations compatible with the tensor dtype and
    head dim, and according to the rank of the token layout (1D/2D/3D).

    Each configuration for this operation is a tuple of two integer tuples: `(q_tile_shape,
    kv_tile_shape)`. These are arguments to [natten.na1d][natten.na1d], [natten.na2d][natten.na2d],
    and [natten.na3d][natten.na3d].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[tuple, tuple]]): List of tuples of two integer tuples corresponding to query
            and KV tile *shapes*.
    """
    if not can_run_cutlass_blackwell_fna(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_blackwell_fna_backward_configs(input_tensor=query)


### CUTLASS Hopper kernels


def get_configs_for_cutlass_hopper_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassHopperFmhaForwardConfigType]:
    """Returns Hopper FMHA configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Hopper GPU (SM90; compute capability 9.0), and if so,
    returns *forward pass* configurations compatible with the tensor dtype and head dim.

    Each configuration for this operation is a tuple of one integer tuple, and another integer:
    `((q_tile_size, kv_tile_size), kernel_schedule)`. These are arguments to
    [natten.attention][natten.attention].
    `kernel_schedule` is specific to Hopper FNA/FMHA only.

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[Tuple[int, int], KernelSchedule]]): List of tuples of one tuple of two integers
            corresponding to query and KV tile sizes, and a kernel schedule enum type.
    """
    if not can_run_cutlass_hopper_fmha(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_hopper_fmha_forward_configs(input_tensor=query)


def get_bwd_configs_for_cutlass_hopper_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassHopperFmhaBackwardConfigType]:
    """Returns Hopper FMHA backward pass configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Hopper GPU (SM90; compute capability 9.0), and if so,
    returns *backward pass* configurations compatible with the tensor dtype and head dim.

    Each configuration for this operation is an integer tuple:
    `((backward_q_tile_size, backward_kv_tile_size), kernel_schedule)`. These are arguments to
    [natten.attention][natten.attention].

    Note that unlike forward pass, kernel schedule is not part of the configuration. All backward
    pass kernels are persistent warp-specialized.
    See CUTLASS's [example 88](https://github.com/NVIDIA/cutlass/tree/main/examples/88_hopper_fmha)
    for more.

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[int, int]]): List of integer tuples corresponding to query and KV tile sizes.
    """
    if not can_run_cutlass_hopper_fmha(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_hopper_fmha_backward_configs(input_tensor=query)


def get_configs_for_cutlass_hopper_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassHopperFnaForwardConfigType]:
    """Returns Hopper FNA configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Hopper GPU (SM90; compute capability 9.0), and if so,
    returns *forward pass* configurations compatible with the tensor dtype and head dim.

    Each configuration for this operation is a tuple of one tuple, and another integer:
    `((q_tile_shape, kv_tile_shape), kernel_schedule)`. These are arguments to
    [natten.na1d][natten.na1d], [natten.na2d][natten.na2d], and [natten.na3d][natten.na3d].

    `kernel_schedule` is specific to Hopper FNA/FMHA only.

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[Tuple[tuple, tuple], KernelSchedule]]): List of tuples of one tuple of two
            shape tuples, corresponding to query and KV tile *shapes*, and a kernel schedule enum
            type.
    """
    if not can_run_cutlass_hopper_fna(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_hopper_fna_forward_configs(input_tensor=query)


def get_bwd_configs_for_cutlass_hopper_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassHopperFnaBackwardConfigType]:
    """Returns Hopper FNA backward pass configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a Hopper GPU (SM90; compute capability 9.0), and if so,
    returns *backward pass* configurations compatible with the tensor dtype and head dim.

    Each configuration for this operation is a tuple of two tuples:
    `(q_tile_shape, kv_tile_shape)`. These are arguments to [natten.na1d][natten.na1d],
    [natten.na2d][natten.na2d], and [natten.na3d][natten.na3d].

    Note that unlike forward pass, kernel schedule is not part of the configuration. All backward
    pass kernels are persistent warp-specialized.
    See CUTLASS's [example 88](https://github.com/NVIDIA/cutlass/tree/main/examples/88_hopper_fmha)
    for more.

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[tuple, tuple]]): List of tuples of two shape tuples, corresponding to query and
        KV tile *shapes*.
    """
    if not can_run_cutlass_hopper_fna(
        query=query, key=key, value=value, raise_error=False
    ):
        return []

    return get_all_hopper_fna_backward_configs(input_tensor=query)


### CUTLASS 2.X kernels


def get_configs_for_cutlass_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassFmhaForwardConfigType]:
    """Returns CUTLASS FMHA configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a device with compute capability >= 5.0, and if so,
    returns *forward pass* configurations compatible with the specific compute capability, tensor
    dtype and head dim.

    Each configuration for this operation is a tuple of two integers: `(q_tile_size,
    kv_tile_size)`. These are arguments to [natten.attention][natten.attention].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[int, int]]): List of tuples of two integers corresponding to query and KV tile
            sizes.
    """
    if not can_run_cutlass_fmha(query=query, key=key, value=value, raise_error=False):
        return []

    return get_all_cutlass_fmha_forward_configs(input_tensor=query)


def get_bwd_configs_for_cutlass_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassFmhaForwardConfigType]:
    """Returns CUTLASS FMHA backward pass configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a device with compute capability >= 5.0, and if so,
    returns *backward pass* configurations compatible with the specific compute capability, tensor
    dtype and head dim.

    Each configuration for this operation is a tuple of two integers: `(backward_q_tile_size,
    backward_kv_tile_size)`. These are arguments to [natten.attention][natten.attention].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[int, int]]): List of tuples of two integers corresponding to query and KV tile
            sizes in the *backward pass*.
    """
    if not can_run_cutlass_fmha(query=query, key=key, value=value, raise_error=False):
        return []

    return get_all_cutlass_fmha_backward_configs(
        input_tensor=key if key.shape[-1] >= value.shape[-1] else value
    )


def get_configs_for_cutlass_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassFnaForwardConfigType]:
    """Returns CUTLASS FNA configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a device with compute capability >= 5.0, and if so,
    returns *forward pass* configurations compatible with the specific compute capability, tensor
    dtype and head dim, and according to the rank of the token layout (1D/2D/3D).

    Each configuration for this operation is a tuple of two integer tuples: `(q_tile_shape,
    kv_tile_shape)`. These are arguments to [natten.na1d][natten.na1d], [natten.na2d][natten.na2d],
    and [natten.na3d][natten.na3d].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[tuple, tuple]]): List of tuples of two integer tuples corresponding to query
            and KV tile *shapes*.
    """
    if not can_run_cutlass_fna(query=query, key=key, value=value, raise_error=False):
        return []

    return get_all_cutlass_fna_forward_configs(input_tensor=query)


def get_bwd_configs_for_cutlass_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
) -> List[CutlassFnaForwardConfigType]:
    """Returns CUTLASS FNA backward pass configurations compatible with input tensors, if any.

    Checks first if a CUDA tensor, and on a device with compute capability >= 5.0, and if so,
    returns *backward pass* configurations compatible with the specific compute capability, tensor
    dtype and head dim.

    Each configuration for this operation is a tuple of two integers: `(backward_q_tile_shape,
    backward_kv_tile_shape)`. These are arguments to [natten.na1d][natten.na1d],
    [natten.na2d][natten.na2d], and [natten.na3d][natten.na3d].

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.

    Returns:
        (List[Tuple[tuple, tuple]]): List of tuples of two integer tuples corresponding to query
            and KV tile *shapes* in the *backward pass*.
    """
    if not can_run_cutlass_fna(query=query, key=key, value=value, raise_error=False):
        return []

    return get_all_cutlass_fna_backward_configs(
        input_tensor=key if key.shape[-1] >= value.shape[-1] else value
    )


### Flex


def get_configs_for_flex_fmha(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    torch_compile: bool = False,
) -> List[FlexFmhaForwardConfigType]:
    """Returns Flex FMHA configurations compatible with input tensors, if any.

    Each configuration for this operation is a tuple of two integers: `(q_tile_size,
    kv_tile_size)`. These are arguments to [natten.attention][natten.attention].
    Not specifying these arguments while backend is Flex will default to `q_tile_size = 64` and
    `kv_tile_size = 64`.

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.
        torch_compile: Whether or not you intend to use compiled block mask and flex attention kernel.

    Returns:
        (List[Tuple[int, int]]): List of tuples of two integers corresponding to query and KV tile
            sizes.
    """
    if not can_run_flex_attention(
        query=query,
        key=key,
        value=value,
        torch_compile=torch_compile,
        raise_error=False,
    ):
        return []

    return get_all_flex_fmha_forward_configs(input_tensor=query)


def get_configs_for_flex_fna(
    query: Tensor,
    key: Tensor,
    value: Tensor,
    torch_compile: bool = False,
) -> List[FlexFnaForwardConfigType]:
    """Returns Flex FNA configurations compatible with input tensors, if any.

    Each configuration for this operation is a tuple of two integer tuples: `(q_tile_shape,
    kv_tile_shape)`. These are arguments to [natten.na1d][natten.na1d], [natten.na2d][natten.na2d],
    and [natten.na3d][natten.na3d].
    Not specifying these arguments while backend is Flex will default to single-dimensional tiling,
    and will not use our Token Permutation approach. By explicitly specifying tile shapes, you will
    automatically use our Token Permutation approach, which saves you the most compute.

    Args:
        query: Query tensor matching the shape, dtype, and device of your use case.
        key:   Key tensor matching the shape, dtype, and device of your use case.
        value: Value tensor matching the shape, dtype, and device of your use case.
        torch_compile: Whether or not you intend to use compiled block mask and flex attention kernel.

    Returns:
        (List[Tuple[tuple, tuple]]): List of tuples of two integer tuples corresponding to query
            and KV tile *shapes*.
    """
    if not can_run_flex_attention(
        query=query,
        key=key,
        value=value,
        torch_compile=torch_compile,
        raise_error=False,
    ):
        return []

    return get_all_flex_fna_forward_configs(input_tensor=query)



#################################################################################################
# Copyright (c) 2022-2025 Ali Hassani.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################################

from typing import Optional

import torch


def is_cuda(device: torch.device) -> bool:
    return torch.cuda.is_available() and torch.version.cuda and device.type == "cuda"  # type: ignore


def is_rocm(device: torch.device) -> bool:
    return torch.cuda.is_available() and torch.version.hip and device.type == "cuda"  # type: ignore


def is_cpu(device: torch.device) -> bool:
    return device.type == "cpu"


def get_device_cc(device: Optional[torch.device] = None) -> int:
    if (
        torch.cuda.is_available()
        and torch.version.cuda
        and (device is None or is_cuda(device))
    ):
        major, minor = torch.cuda.get_device_capability(device)
        return major * 10 + minor

    return 0
